"""Baseclass for all tests."""
import copy
import inspect
import logging
import os
import sys
import time
import unittest

from mpf.core.config_loader import YamlMultifileConfigLoader

from unittest.mock import *

import asyncio
from asyncio import events

from mpf.core.logging import LogMixin
from mpf.core.rgb_color import RGBColor

from mpf.tests.TestDataManager import TestDataManager
from mpf.tests.loop import TimeTravelLoop, TestClock

import mpf.core
import mpf.core.config_validator
from mpf.core.machine import MachineController
from mpf.core.utility_functions import Util
from mpf.file_interfaces.yaml_interface import YamlInterface

YamlInterface.cache = True
UNITTEST_CONFIG_CACHE = {}      # type: ignore


class UnitTestConfigLoader(YamlMultifileConfigLoader):

    """Config Loader for Unit Tests."""

    def __init__(self, machine_path, configfile, config_defaults, config_patches, spec_patches):
        super().__init__(machine_path, configfile, True, True)
        self.config_defaults = config_defaults
        self.config_patches = config_patches
        self.spec_patches = spec_patches

    def _load_config_spec(self):
        config_spec = super()._load_config_spec()
        if self.spec_patches:
            config_spec = Util.dict_merge(config_spec, self.spec_patches)
        return config_spec

    def _load_mpf_machine_config(self, config_spec):
        config = super()._load_mpf_machine_config(config_spec)

        if self.config_defaults:
            config = Util.dict_merge(self.config_defaults, config)
        if self.config_patches:
            config = Util.dict_merge(config, self.config_patches, False)
        return config


class MpfUnitTestFormatter(logging.Formatter):

    """Log Test and Real Time."""

    def formatTime(self, record, datefmt=None):
        """Return Real Time and Test Time."""
        return "{} : {:.3f}".format(super().formatTime(record, datefmt), record.created_clock)


def test_config(config_file):

    """Decorator to overwrite config file for one test."""

    def test_decorator(fn):
        """Decorate function."""
        fn.config_file = config_file
        return fn
    return test_decorator


def test_config_directory(config_directory):

    """Decorator to overwrite config directory for one test."""

    def test_decorator(fn):
        """Decorate function."""
        fn.config_directory = config_directory
        return fn
    return test_decorator


class TestMachineController(MachineController):

    """A patched version of the MachineController used in tests.

    The TestMachineController has a few changes from the regular machine
    controller to facilitate running unit tests, including:

    * Use the TestDataManager instead of the real one.
    * Use a test clock which we can manually advance instead of the regular
      clock tied to real-world time.
    * Only load plugins if ``self._enable_plugins`` is *True*.
    * Merge any ``test_config_patches`` into the machine config.
    * Disabled the config file caching to always load the config from disk.

    """

    def __init__(self, options, config, config_patches, config_defaults, clock, mock_data,
                 enable_plugins=False):
        self.test_config_patches = config_patches
        self.test_config_defaults = config_defaults
        self._enable_plugins = enable_plugins
        self._test_clock = clock
        self._mock_data = mock_data
        super().__init__(options, config)

    def create_data_manager(self, config_name):
        """Create TestDataManager."""
        return TestDataManager(self._mock_data.get(config_name, {}))

    def _load_clock(self):
        return self._test_clock

    def __del__(self):
        if self._test_clock:
            self._test_clock.loop.close()

    def _register_plugin_config_players(self):
        if self._enable_plugins:
            super()._register_plugin_config_players()


class MpfTestCase(unittest.TestCase):

    """Primary TestCase class used for all MPF unit tests."""

    def __init__(self, methodName='runTest'):
        LogMixin.unit_test = True

        super().__init__(methodName)
        self.machine = None     # type: TestMachineController
        self.machine_config_patches = dict()
        self.machine_config_patches['mpf'] = dict()
        self.machine_config_patches['mpf']['default_platform_hz'] = 100
        self.machine_config_patches['mpf']['plugins'] = list()
        self.machine_config_patches['bcp'] = []

        self.machine_config_defaults = dict()
        self.machine_config_defaults['playfields'] = dict()
        self.machine_config_defaults['playfields']['playfield'] = dict()
        self.machine_config_defaults['playfields']['playfield']['tags'] = "default"
        self.machine_config_defaults['playfields']['playfield']['default_source_device'] = None

        self.machine_spec_patches = {}

        self._last_event_kwargs = {}
        self._events = {}
        self.expected_duration = 0.5

    def start_mode(self, mode):
        """Start mode."""
        self.assertIn(mode, self.machine.modes)
        self.assertModeNotRunning(mode)
        self.machine.modes[mode].start()
        self.machine_run()
        self.assertModeRunning(mode)

    def stop_mode(self, mode):
        """Stop mode."""
        self.assertModeRunning(mode)
        self.machine.modes[mode].stop()
        self.machine_run()
        self.assertModeNotRunning(mode)

    def get_config_file(self):
        """Return a string name of the machine config file to use for the tests
        in this class.

        You should override this method in your own test class to point to the
        config file you need for your tests.

        Returns:
            A string name of the machine config file to use, complete with the
            ``.yaml`` file extension.

        For example:

        .. code::

            def get_config_file(self):
                return 'my_config.yaml'

        """
        return 'null.yaml'

    def get_machine_path(self):
        """Return a string name of the path to the machine folder to use for
        the tests in this class.

        You should override this method in your own test class to point to the
        machine folder root you need for your tests.

        Returns:
            A string name of the machine path to use

        For example:

        .. code::

            def get_machine_path(self):
                return 'tests/machine_files/my_test/'

        Note that this path is relative to the MPF package root

        """
        return 'tests/machine_files/null/'

    def get_absolute_machine_path(self):
        """Return absolute machine path."""
        # check if there is a decorator
        config_directory = getattr(getattr(self, self._testMethodName), "config_directory", None)
        if not config_directory:
            config_directory = self.get_machine_path()

        # creates an absolute path based on machine_path
        return os.path.abspath(os.path.join(
            mpf.core.__path__[0], os.pardir, config_directory))

    @staticmethod
    def get_abs_path(path):
        """Get absolute path relative to current directory."""
        return os.path.join(os.path.abspath(os.curdir), path)

    def post_event(self, event_name, run_time=0):
        """Post an MPF event and optionally advance the time.

        Args:
            event_name: String name of the event to post
            run_time: How much time (in seconds) the test should advance
                after this event has been posted.

        For example, to post an event called "shot1_hit":

        .. code::

            self.post_event('shot1_hit')

        To post an event called "tilt" and then advance the time 1.5 seconds:

        .. code::

            self.post_event('tilt', 1.5)

        """
        self.machine.events.post(event_name)
        self.advance_time_and_run(run_time)

    def post_event_with_params(self, event_name, **params):
        """Post an MPF event with kwarg parameters.

        Args:
            event_name: String name of the event to post
            **params: One or more kwarg key/value pairs to post with the event.

        For example, to post an event called "jackpot" with the parameters
        ``count=1`` and ``first_time=True``, you would use:

        .. code::

            self.post_event('jackpot', count=1, first_time=True)

        """
        self.machine.events.post(event_name, **params)
        self.machine_run()

    def post_relay_event_with_params(self, event_name, **params):
        """Post a relay event synchronously and return the result."""
        future = self.machine.events.post_relay_async(event_name, **params)
        result = self.machine.clock.loop.run_until_complete(future)
        return result

    def assertPlaceholderEvaluates(self, expected, condition):
        """Assert that a placeholder evaluates to a certain value."""
        result = self.machine.placeholder_manager.build_raw_template(condition).evaluate([],
                                                                                         fail_on_missing_params=True)
        self.assertEqual(expected, result, "{} = {} != {}".format(condition, result, expected))

    def assertNumBallsKnown(self, balls):
        """Assert that a certain number of balls are known in the machine."""
        self.assertEqual(balls, self.machine.ball_controller.num_balls_known)

    def set_num_balls_known(self, balls):
        """Set the ball controller's ``num_balls_known`` attribute.

        This is needed for tests where you don't have any ball devices and
        other situations where you need the ball controller to think the
        machine has a certain amount of balls to run a test.

        Example:

        .. code::

            self.set_num_balls_known(3)


        """
        # in case the test does not have any ball devices
        self.machine.ball_controller.num_balls_known = balls

    def get_platform(self):
        """Force this test class to use a certain platform.

        Returns:
            String name of the platform this test class will use.

        If you don't include this method in your test class, the platform will
        be set to `virtual`. If you want to use the smart virtual platform,
        you would add the following to your test class:

        .. code::

            def get_platform(self):
                return 'smart_virtual`

        """
        return 'virtual'

    def get_use_bcp(self):
        """Control whether tests in this class should use BCP.

        Returns: True or False

        The default is False. To use BCP in your test class, add the following:

         .. code::

            def get_use_bcp(self):
                return True

        """
        return False

    def get_enable_plugins(self):
        """Control whether tests in this class should load MPF plugins.

        Returns: True or False

        The default is False. To load plugins in your test class, add the
        following:

         .. code::

            def get_enable_plugins(self):
                return True

        """
        return False

    def _get_config_file(self):
        """Return test decorator value or the return of get_config_file."""
        config_file = getattr(getattr(self, self._testMethodName), "config_file", None)
        if config_file:
            return config_file
        else:
            return self.get_config_file()

    def get_options(self):
        """Return options for the machine controller."""
        mpfconfig = os.path.abspath(os.path.join(
            mpf.core.__path__[0], os.pardir, 'mpfconfig.yaml'))

        return {
            'force_platform': self.get_platform(),
            'production': False,
            'mpfconfigfile': mpfconfig,
            'configfile': Util.string_to_event_list(self._get_config_file()),
            'debug': True,
            'bcp': self.get_use_bcp(),
            'no_load_cache': False,
            'create_config_cache': True,
            'text_ui': False,
        }

    def advance_time_and_run(self, delta=1.0):
        """Advance the test clock and run anything that should run during that
        time.

        Args:
            delta: How much time to advance the test clock by (in seconds)

        This method will cause anything scheduled during the time to run,
        including things like delays, timers, etc.

        Advancing the clock will happen in multiple small steps if things are
        scheduled to happen during this advance. For example, you can advance
        the clock 10 seconds like this:

        .. code::

            self.advance_time_and_run(10)

        If there is a delay callback that is scheduled to happen in 2 seconds,
        then this method will advance the clock 2 seconds, process that delay,
        and then advance the remaining 8 seconds.

        """
        self.machine.log.info("Advancing time by %s", delta)
        try:
            self.loop.run_until_complete(asyncio.sleep(delay=delta))
            return
        except RuntimeError as e:
            try:
                self.machine.stop()
            except Exception:
                pass
            if self._exception and "exception" in self._exception:
                exception = self._exception['exception']
                self._exception = None
                raise exception
            elif self._exception:
                exception = self._exception['exception']
                self._exception = None
                raise Exception(exception, e)
            raise e

    def machine_run(self):
        """Process any delays, timers, or anything else scheduled.

        Note this is the same as:

        .. code::

            self.advance_time_and_run(0)

        """
        self.advance_time_and_run(0)

    @staticmethod
    def unittest_verbosity():
        """Return the verbosity setting of the currently running unittest
        program, or 0 if none is running.

        Returns: An integer value of the current verbosity setting.

        """
        frame = inspect.currentframe()
        while frame:
            obj = frame.f_locals.get('self')
            if isinstance(obj, unittest.TestProgram) or isinstance(obj,
                                                                   unittest.TextTestRunner):
                return obj.verbosity
            frame = frame.f_back
        return 0

    def save_and_prepare_sys_path(self):
        """Save sys path and prepare it for the test."""
        # save path
        self._sys_path = copy.deepcopy(sys.path)

        mpf_path = os.path.abspath(os.path.join(mpf.core.__path__[0], os.pardir))
        if mpf_path in sys.path:
            sys.path.remove(mpf_path)

        # make tests path independent. remove current dir absolue
        if os.curdir in sys.path:
            sys.path.remove(os.curdir)

        # make tests path independent. remove current dir relative
        if "" in sys.path:
            sys.path.remove("")

    def restore_sys_path(self):
        """Restore sys path after test."""
        sys.path = self._sys_path

    def _get_mock_data(self):
        """Return data for MockDataMangager in test."""
        return dict()

    def _mock_loop(self):
        pass

    def _early_machine_init(self, machine):
        pass

    def _exception_handler(self, loop, context):
        try:
            loop.stop()
        except RuntimeError:
            pass

        self._exception = context

    def setUp(self):
        """Setup test."""
        # we want to reuse config_specs to speed tests up
        mpf.core.config_validator.ConfigValidator.unload_config_spec = (
            MagicMock())

        self._events = {}
        self._last_event_kwargs = {}
        self._exception = None

        # print(threading.active_count())

        self.test_start_time = time.time()
        self.save_and_prepare_sys_path()

        # init machine
        machine_path = self.get_absolute_machine_path()

        self.loop = TimeTravelLoop()
        events.set_event_loop(self.loop)
        self.loop.set_exception_handler(self._exception_handler)
        self.clock = TestClock(self.loop)
        self._mock_loop()

        if self.unittest_verbosity() > 1:
            class LogRecordClock(logging.LogRecord):
                def __init__(self_inner, *args, **kwargs):
                    self_inner.created_clock = self.clock.get_time()
                    super().__init__(*args, **kwargs)

            logging.setLogRecordFactory(LogRecordClock)

            console_log = logging.StreamHandler()
            console_log.setLevel(logging.DEBUG)
            console_log.setFormatter(MpfUnitTestFormatter(fmt='%(asctime)s : %(levelname)s : %('
                                                              'name)s : %(message)s'))
            logging.basicConfig(level=logging.DEBUG,
                                handlers=[console_log])
        else:
            # no logging by default
            logging.basicConfig(level=99)

        # load config
        config_loader = UnitTestConfigLoader(machine_path, [self._get_config_file()], self.machine_config_defaults,
                                             self.machine_config_patches, self.machine_spec_patches)

        config = config_loader.load_mpf_config()

        try:
            self.machine = TestMachineController(
                self.get_options(), config, self.machine_config_patches, self.machine_config_defaults,
                self.clock, self._get_mock_data(), self.get_enable_plugins())

            self._early_machine_init(self.machine)

            self._initialise_machine()

        # pylint: disable-msg=broad-except
        except Exception as e:
            try:
                self.machine.stop()
            except AttributeError:
                pass
            if self._exception and 'exception' in self._exception:
                raise self._exception['exception']
            elif self._exception:
                raise Exception(self._exception, e)
            raise e

    def _initialise_machine(self):
        init = asyncio.ensure_future(self.machine.initialise())
        self._wait_for_start(init, 20)
        self.machine.events.process_event_queue()
        self.advance_time_and_run(.001)

    def _wait_for_start(self, init, timeout):
        start = time.time()
        while not init.done() and not self._exception:
            self.loop.run_once()
            if time.time() > start + timeout:
                raise AssertionError("Start took more than {}s".format(timeout))

        # trigger exception if there was one
        init.result()

    def _mock_event_handler(self, event_name, **kwargs):
        self._last_event_kwargs[event_name] = kwargs
        self._events[event_name] += 1

    def mock_event(self, event_name):
        """Configure an event to be mocked.

        Args:
            event_name: String name of the event to mock.

        Mocking an event is an easy way to check if an event was called without
        configuring some kind of callback action in your tests.

        Note that an event must be mocked here *before* it's posted in order
        for :meth:`assertEventNotCalled` and :meth:`assertEventCalled` to work.

        Mocking an event will not "break" it. In other words, any other
        registered handlers for this event will also be called even if the
        event has been mocked.

        For example:

        .. code::

            self.mock_event('my_event')
            self.assertEventNotCalled('my_event')  # This will be True
            self.post_event('my_event')
            self.assertEventCalled('my_event')  # This will also be True

        """
        self._events[event_name] = 0
        self._last_event_kwargs.pop(event_name, None)
        self.machine.events.remove_handler_by_event(event=event_name, handler=self._mock_event_handler)
        self.machine.events.add_handler(event=event_name,
                                        handler=self._mock_event_handler,
                                        event_name=event_name)

    def assertBallsOnPlayfield(self, balls, playfield="playfield"):
        """Assert that a certain number of ball is on a playfield."""
        self.assertEqual(balls, self.machine.playfields[playfield].balls,
                         "Playfield {} contains {} balls but should be {}".format(
                             playfield, self.machine.playfields[playfield].balls, balls))

    def assertAvailableBallsOnPlayfield(self, balls, playfield="playfield"):
        """Assert that a certain number of ball is available on a playfield."""
        self.assertEqual(balls, self.machine.playfields[playfield].available_balls)

    def assertMachineVarEqual(self, value, machine_var):
        """Assert that a machine variable exists and has a certain value."""
        self.assertTrue(self.machine.variables.is_machine_var(machine_var), "Machine Var {} does not exist.".format(machine_var))
        self.assertEqual(value, self.machine.variables.get_machine_var(machine_var))

    def assertPlayerVarEqual(self, value, player_var):
        """Assert that a player variable exists and has a certain value."""
        self.assertIsNotNone(self.machine.game, "There is no game.")
        self.assertEqual(value, self.machine.game.player[player_var], "Value of player var {} is {} but should be {}".
                         format(player_var, self.machine.game.player[player_var], value))

    def assertSwitchState(self, name, state):
        """Assert that a switch exists and has a certain state."""
        self.assertIn(name, self.machine.switches, "Switch {} does not exist.".format(name))
        self.assertEqual(state, self.machine.switch_controller.is_active(self.machine.switches[name]),
                         "Switch {} is in state {} != {}".format(
                             name, self.machine.switch_controller.is_active(self.machine.switches[name]), state))

    def assertLightChannel(self, light_name, brightness, channel="white"):
        """Assert that a light channel has a certain brightness."""
        self.assertIn(light_name, self.machine.lights, "Light {} does not exist".format(light_name))
        self.assertAlmostEqual(brightness / 255.0, self.machine.lights[light_name].hw_drivers[channel][0].
                               current_brightness)

    def assertNotLightChannel(self, light_name, brightness, channel="white"):
        """Assert that a light channel does not have a certain brightness."""
        self.assertIn(light_name, self.machine.lights, "Light {} does not exist".format(light_name))
        self.assertNotEqual(brightness, self.machine.lights[light_name].hw_drivers[channel][0].
                            current_brightness)

    def assertLightColor(self, light_name, color):
        """Assert that a light exists and shows one color."""
        self.assertIn(light_name, self.machine.lights, "Light {} does not exist".format(light_name))
        if isinstance(color, str) and color.lower() == 'on':
            color = self.machine.lights[light_name].config['default_on_color']

        self.assertEqual(RGBColor(color), self.machine.lights[light_name].get_color(),
                         "Light {}: {} != {}. Stack: {}".format(
                             light_name, RGBColor(color).name, self.machine.lights[light_name].get_color().name,
                             self.machine.lights[light_name].stack))

    def assertNotLightColor(self, light_name, color):
        """Assert that a light exists and does not show a certain color."""
        self.assertIn(light_name, self.machine.lights, "Light {} does not exist".format(light_name))
        if isinstance(color, str) and color.lower() == 'on':
            color = self.machine.lights[light_name].config['default_on_color']

        self.assertNotEqual(RGBColor(color), self.machine.lights[light_name].get_color(),
                            "{} == {}".format(RGBColor(color).name, self.machine.lights[light_name].get_color().name))

    def assertLightColors(self, light_name, color_list, secs=1, check_delta=.1):
        """Assert that a light exists and shows all colors in a list over a period."""
        self.assertIn(light_name, self.machine.lights, "Light {} does not exist".format(light_name))
        colors = list()

        # have to do it this weird way because `if 'on' in color_list:` doesn't
        # work since it tries to convert it to a color
        for color in color_list[:]:
            if isinstance(color, str) and color.lower() == 'on':
                color_list.remove('on')
                color_list.append(self.machine.lights[light_name].config['default_on_color'])
                break

        for x in range(int(secs / check_delta)):
            color = self.machine.lights[light_name].get_color()
            colors.append(color)
            self.advance_time_and_run(check_delta)

        for color in color_list:
            self.assertIn(RGBColor(color), colors)

    def assertLightOn(self, light_name):
        """Assert that a light exists and is on."""
        self.assertIn(light_name, self.machine.lights, "Light {} does not exist".format(light_name))
        self.assertEqual(255,
                         self.machine.lights[
                             light_name].hw_driver.current_brightness)

    def assertLightOff(self, light_name):
        """Assert that a light exists and is off."""
        self.assertIn(light_name, self.machine.lights, "Light {} does not exist".format(light_name))
        self.assertEqual(0, self.machine.lights[light_name].hw_driver.current_brightness)

    def assertLightFlashing(self, light_name, color=None, secs=1, check_delta=.1):
        """Assert that a light exists and is flashing."""
        self.assertIn(light_name, self.machine.lights, "Light {} does not exist".format(light_name))
        brightness_values = list()
        if not color:
            color = [255, 255, 255]

        for _ in range(int(secs / check_delta)):
            brightness_values.append(
                self.machine.lights[light_name].get_color())
            self.advance_time_and_run(check_delta)

        self.assertIn([0, 0, 0], brightness_values)
        self.assertIn(color, brightness_values)

    def assertModeRunning(self, mode_name):
        """Assert that a mode exists and is running."""
        if mode_name not in self.machine.modes:
            raise AssertionError("Mode {} not known.".format(mode_name))
        self.assertIn(self.machine.modes[mode_name], self.machine.mode_controller.active_modes,
                      "Mode {} not running.".format(mode_name))

    def assertModeNotRunning(self, mode_name):
        """Assert that a mode exists and is not running."""
        if mode_name not in self.machine.modes:
            raise AssertionError("Mode {} not known.".format(mode_name))
        self.assertNotIn(self.machine.modes[mode_name], self.machine.mode_controller.active_modes,
                         "Mode {} running but should not.".format(mode_name))

    def assertEventNotCalled(self, event_name):
        """Assert that event was not called.

        Args:
            event_name: String name of the event to check.

        Note that the event must be mocked via ``self.mock_event()`` first in
        order to use this method.

        """
        if event_name not in self._events:
            raise AssertionError("Event {} not mocked.".format(event_name))

        if self._events[event_name] != 0:
            raise AssertionError("Event {} was called {} times but was not expected to "
                                 "be called.".format(event_name, self._events[event_name]))

    def assertEventCalled(self, event_name, times=None):
        """Assert that event was called.

        Args:
            event_name: String name of the event to check.
            times: An optional value to confirm the number of times the event
                was called. Default of *None* means this method will pass as
                long as the event has been called at least once.

        If you want to reset the ``times`` count, you can mock the event
        again.

        Note that the event must be mocked via ``self.mock_event()`` first in
        order to use this method.

        For example:

        .. code::

            self.mock_event('my_event')
            self.assertEventNotCalled('my_event')  # This will pass

            self.post_event('my_event')
            self.assertEventCalled('my_event')     # This will pass
            self.assertEventCalled('my_event', 1)  # This will pass

            self.post_event('my_event')
            self.assertEventCalled('my_event')     # This will pass
            self.assertEventCalled('my_event', 2)  # This will pass

        """
        if event_name not in self._events:
            raise AssertionError("Event {} not mocked.".format(event_name))

        if self._events[event_name] == 0 and times != 0:
            if times is None:
                raise AssertionError("Event {} was not called.".format(event_name))

            raise AssertionError("Event {} was not called but we expected {} calls.".format(event_name, times))

        if times is not None and self._events[event_name] != times:
            raise AssertionError("Event {} was called {} times instead of {} times expected.".format(
                event_name, self._events[event_name], times))

    def assertEventCalledWith(self, event_name, **kwargs):
        """Assert an event was called with certain kwargs.

        Args:
            event_name: String name of the event to check.
            **kwargs: Name/value parameters to check.

        For example:

        .. code::

            self.mock_event('jackpot')

            self.post_event('jackpot', count=1, first_time=True)
            self.assertEventCalled('jackpot')  # This will pass
            self.assertEventCalledWith('jackpot', count=1, first_time=True)  # This will also pass
            self.assertEventCalledWith('jackpot', count=1, first_time=False)  # This will fail


        """
        self.assertEventCalled(event_name)
        self.assertEqual(kwargs, self._last_event_kwargs[event_name], "Args for {} differ.".format(event_name))

    def assertColorAlmostEqual(self, color1, color2, delta=6):
        """Assert that two color are almost equal.

        Args:
            color1: The first color, as an RGBColor instance or 3-item iterable.
            color2: The second color, as an RGBColor instance or 3-item iterable.
            delta: How close the colors have to be. The deltas between red,
                green, and blue are added together and must be less or equal
                to this value for this assertion to succeed.

        """
        if isinstance(color1, RGBColor) and isinstance(color2, RGBColor):
            difference = abs(color1.red - color2.red) +\
                abs(color1.blue - color2.blue) +\
                abs(color1.green - color2.green)
        else:
            difference = abs(color1[0] - color2[0]) +\
                abs(color1[1] - color2[1]) +\
                abs(color1[2] - color2[2])
        self.assertLessEqual(difference, delta, "Colors do not match: " + str(color1) + " " + str(color2))

    def reset_mock_events(self):
        """Reset all mocked events.

        This will reset the count of number of times called every mocked
        event is.

        """
        for event in self._events.keys():
            self._events[event] = 0

    def hit_switch_and_run(self, name, delta):
        """Activates a switch and advances the time.

        Args:
            name: The name of the switch to activate.
            delta: The time (in seconds) to advance the clock.

        Note that this method does not deactivate the switch once the time
        has been advanced, meaning the switch stays active. To make the
        switch inactive, use the :meth:`release_switch_and_run`.

        """
        self.machine.switch_controller.process_switch(name, state=1, logical=True)
        if delta:
            self.advance_time_and_run(delta)

    def release_switch_and_run(self, name, delta):
        """Deactivates a switch and advances the time.

        Args:
            name: The name of the switch to activate.
            delta: The time (in seconds) to advance the clock.

        """
        self.machine.switch_controller.process_switch(name, state=0, logical=True)
        if delta:
            self.advance_time_and_run(delta)

    def hit_and_release_switch(self, name):
        """Momentarily activates and then deactivates a switch.

        Args:
            name: The name of the switch to hit.

        This method immediately activates and deactivates a switch with no
        time in between.

        """
        self.machine.switch_controller.process_switch(name, state=1, logical=True)
        self.machine.switch_controller.process_switch(name, state=0, logical=True)
        self.machine_run()

    def hit_and_release_switches_simultaneously(self, names):
        """Momentarily activates and then deactivates multiple switches.

        Switches are hit sequentially and then released sequentially.
        Events are only processed at the end of the sequence which is useful
        to reproduce race conditions when processing nearly simultaneous hits.

        Args:
            names: The names of the switches to hit and release.

        """
        for name in names:
            self.machine.switch_controller.process_switch(name, state=1, logical=True)
        for name in names:
            self.machine.switch_controller.process_switch(name, state=0, logical=True)
        self.machine_run()

    def tearDown(self):
        """Tear down test."""
        if self._exception:
            try:
                self.machine.shutdown()
            except:
                pass

            if self._exception and 'exception' in self._exception:
                raise self._exception['exception']
            elif self._exception:
                raise Exception(self._exception)

        duration = time.time() - self.test_start_time
        if duration > self.expected_duration:
            print("Test {}.{} took {} > {}s".format(self.__class__,
                  self._testMethodName, round(duration, 2),
                  self.expected_duration))

        self.machine.log.debug("Test ended")
        if sys.exc_info != (None, None, None):
            # disable teardown logging after error
            logging.basicConfig(level=99)
        else:
            # fire all delays
            self.advance_time_and_run(300)
        self.machine._do_stop()
        self.machine = None

        self.restore_sys_path()
        events.set_event_loop(None)

    @staticmethod
    def add_to_config_validator(machine, key, new_dict):
        """Add config dict to validator."""
        machine.config_validator.get_config_spec()[key] = new_dict



"""Contains the base class for the Snux driver overlay.

This class overlays an existing WPC-compatible platform interface to work with
Mark Sunnucks's System 11 interface board.
"""
from typing import Any

from mpf.core.machine import MachineController
from mpf.platforms.system11 import System11OverlayPlatform


class SnuxHardwarePlatform(System11OverlayPlatform):

    """Overlay platform for the snux hardware board."""

    __slots__ = ["snux_config"]

    def __init__(self, machine: MachineController) -> None:
        """Initalise the board."""
        super().__init__(machine)
        self.snux_config = None         # type: Any

    def _null_log_handler(self, *args, **kwargs):
        pass

    def _initialize(self, **kwargs):
        super()._initialize(**kwargs)

        self.log.debug("Configuring Snux Diag LED for driver %s",
                       self.snux_config['diag_led_driver'].name)

        # Hack to silence logging of P_ROC
        # TODO: clean this up
        self.snux_config['diag_led_driver'].hw_driver.log.info = self._null_log_handler
        self.snux_config['diag_led_driver'].hw_driver.log.debug = self._null_log_handler

        self.snux_config['diag_led_driver'].enable()

        self.machine.events.add_handler('init_phase_5',
                                        self._initialize_phase_2)

    def _initialize_phase_2(self, **kwargs):
        del kwargs
        self.machine.clock.schedule_interval(self._flash_diag_led, 0.5)

    def _validate_config(self):
        super()._validate_config()
        self.snux_config = self.machine.config_validator.validate_config(
            'snux', self.machine.config['snux'])

    def _flash_diag_led(self):
        """Flash diagnosis LED."""
        self.snux_config['diag_led_driver'].pulse(250)



from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase
from mpf.tests.MpfTestCase import MagicMock, patch, call


class TestOsc(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/osc/'

    def get_platform(self):
        # no force platform
        return False

    async def _start_serve(self):
        return MagicMock(), None

    def setUp(self):
        client_class = patch('mpf.platforms.osc.SimpleUDPClient')
        self.client_class = client_class.start()
        self.client_instance = MagicMock()
        self.client_class.return_value = self.client_instance

        dispatcher_class = patch('mpf.platforms.osc.Dispatcher')
        self.dispatcher_class = dispatcher_class.start()
        self.dispatcher_instance = MagicMock()
        self.dispatcher_class.return_value = self.dispatcher_instance

        server_class = patch('mpf.platforms.osc.AsyncIOOSCUDPServer')
        self.server_class = server_class.start()
        self.server_instance = MagicMock()
        self.server_instance.create_serve_endpoint = self._start_serve
        self.server_class.return_value = self.server_instance

        self.addCleanup(self.client_class.stop)
        self.addCleanup(self.dispatcher_class.stop)
        self.addCleanup(self.server_class.stop)
        super().setUp()

    def test_osc_platform(self):
        # test lights
        self.client_instance.send_message = MagicMock()
        self.machine.lights["test_light1"].color("red")
        self.advance_time_and_run(.1)
        self.client_instance.send_message.assert_has_calls([
            call('/light/light1/blue', 0.0),
            call('/light/light1/green', 0.0),
            call('/light/light1/red', 1.0)], any_order=True)

        self.client_instance.send_message = MagicMock()
        self.machine.lights["test_light1"].color("blue")
        self.advance_time_and_run(.1)
        self.client_instance.send_message.assert_has_calls([
            call('/light/light1/blue', 1.0),
            call('/light/light1/green', 0.0),
            call('/light/light1/red', 0.0)], any_order=True)

        self.client_instance.send_message = MagicMock()
        self.machine.lights["test_light1"].color("white")
        self.advance_time_and_run(.1)
        self.client_instance.send_message.assert_has_calls([
            call('/light/light1/blue', 1.0),
            call('/light/light1/green', 1.0),
            call('/light/light1/red', 1.0)], any_order=True)

        self.client_instance.send_message = MagicMock()
        self.machine.lights["test_light2"].color("white")
        self.advance_time_and_run(.1)
        self.client_instance.send_message.assert_has_calls([
            call('/light/light2/blue', 1.0),
            call('/light/light2/green', 1.0),
            call('/light/light2/red', 1.0)], any_order=True)

        # send switch hits
        self.assertSwitchState("switch_abc", False)
        self.machine.default_platform._handle_switch("/sw/abc", True)
        self.advance_time_and_run(.1)
        self.assertSwitchState("switch_abc", True)
        self.machine.default_platform._handle_switch("/sw/abc", False)
        self.advance_time_and_run(.1)
        self.assertSwitchState("switch_abc", False)

        self.assertSwitchState("switch_1", False)
        self.machine.default_platform._handle_switch("/sw/1", True)
        self.advance_time_and_run(.1)
        self.assertSwitchState("switch_1", True)
        self.machine.default_platform._handle_switch("/sw/1", False)
        self.advance_time_and_run(.1)
        self.assertSwitchState("switch_1", False)

        # test outgoing events
        self.client_instance.send_message = MagicMock()
        self.post_event_with_params("my_test_event", a=100, b=True)
        self.advance_time_and_run(.1)
        self.client_instance.send_message.assert_has_calls([
            call('/event/my_test_event', ['a', 100, 'b', True])
        ], any_order=True)

        self.client_instance.send_message = MagicMock()
        self.post_event("my_other_test_event")
        self.advance_time_and_run(.1)
        self.client_instance.send_message.assert_has_calls([
            call('/event/my_other_test_event', [])
        ], any_order=True)

        # test incoming events
        self.mock_event("test_event")
        self.machine.default_platform._handle_event("/event/test_event", "a", 200, "b", "asd")
        self.advance_time_and_run(.1)
        self.assertEventCalledWith("test_event", a=200, b="asd")

        self.start_game()
        self.advance_time_and_run()

        self.client_instance.send_message.assert_has_calls([
            call('/event/player_turn_started', ['number', 1, 'player', '<Player 1>'])
        ], any_order=True)



"""Contains the YamlInterface class for reading & writing YAML files.

Fixes for octal and boolean values are from here:
http://stackoverflow.com/questions/32965846/cant-parse-yaml-correctly/
"""
import copy
import re
from typing import Any, Iterable, List
from typing import Dict

from collections.abc import Hashable

from ruamel import yaml
from ruamel.yaml import MappingNode
from ruamel.yaml.error import MarkedYAMLError
from ruamel.yaml.reader import Reader
from ruamel.yaml.resolver import BaseResolver
from ruamel.yaml.scanner import Scanner
from ruamel.yaml.parser import Parser
from ruamel.yaml.composer import Composer
from ruamel.yaml.constructor import Constructor, ConstructorError

from mpf.core.file_interface import FileInterface


class MpfResolver(BaseResolver):

    """Resolver with mentioned fixes."""

    @property
    def processing_version(self):
        """Return yaml version."""
        return (1, 2)


RESOLVERS = [
    (
        # Process any item beginning with a plus sign (+) as a string
        u'tag:yaml.org,2002:str',
        re.compile(
            u'''^(\\+([0-9a-zA-Z .]+))$''',
            re.X),
        list(u'+')),

    (
        # Process any 3+ digit number with a leading zero as a string
        u'tag:yaml.org,2002:str',
        re.compile(
            u'''^(?:(0[0-9]{2,}))$''',
            re.X),
        list(u'0')),

    (
        u'tag:yaml.org,2002:bool',
        re.compile(
            u'''^(?:true|True|TRUE|false|False|FALSE|yes|Yes|YES|no|No|NO)$''',
            re.X), list(u'tTfFyYnN')),

    (
        u'tag:yaml.org,2002:float',
        re.compile(u'''^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*
    |\\.[0-9_]+)$''', re.X),
        list(u'-+0123456789.')),

    (
        u'tag:yaml.org,2002:int',
        re.compile(u'''^(?:[-+]?0b[0-1_]+
    |[-+]?[0-9]+
    |[-+]?0o?[0-7_]+
    |[-+]?0x[0-9a-fA-F_]+)$''', re.X),
        list(u'-+0123456789')),

    (
        u'tag:yaml.org,2002:merge',
        re.compile(u'^(?:<<)$'),
        [u'<']),

    (
        u'tag:yaml.org,2002:null',
        re.compile(u'''^(?: ~
    |null|Null|NULL
    | )$''', re.X),
        [u'~', u'n', u'N', u'']),

    (
        u'tag:yaml.org,2002:timestamp',
        re.compile(u'''^(?:[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]
    |[0-9][0-9][0-9][0-9] -[0-9][0-9]? -[0-9][0-9]?
    (?:[Tt]|[ \\t]+)[0-9][0-9]?
    :[0-9][0-9] :[0-9][0-9] (?:\\.[0-9]*)?
    (?:[ \\t]*(?:Z|[-+][0-9][0-9]?(?::[0-9][0-9])?))?)$''', re.X),
        list(u'0123456789')),

    (
        u'tag:yaml.org,2002:value',
        re.compile(u'^(?:=)$'),
        [u'=']),

    # The following resolver is only for documentation purposes. It cannot work
    # because plain scalars cannot start with '!', '&', or '*'.
    (
        u'tag:yaml.org,2002:yaml',
        re.compile(u'^(?:!|&|\\*)$'),
        list(u'!&*'))
]


for resolver in RESOLVERS:
    MpfResolver.add_implicit_resolver(*resolver)


class MpfConstructor(Constructor):

    """Constructor with fix."""

    def construct_mapping(self, node, deep=False):
        """Construct mapping but raise error when a section is defined twice.

        From: http://stackoverflow.com/questions/34358974/how-to-prevent-re-definition-of-keys-in-yaml
        """
        if not isinstance(node, MappingNode):  # pragma: no cover
            raise ConstructorError(problem="expected a mapping node, but found %s" % node.id,
                                   problem_mark=node.start_mark)
        mapping = {}
        for key_node, value_node in node.value:
            # keys can be list -> deep
            key = self.construct_object(key_node, deep=True)
            # lists are not hashable, but tuples are
            if not isinstance(key, Hashable):   # pragma: no cover
                if isinstance(key, list):
                    key = tuple(key)
            if not isinstance(key, Hashable):   # pragma: no cover
                raise ConstructorError(
                    "while constructing a mapping", node.start_mark,
                    "found unhashable key", key_node.start_mark)

            value = self.construct_object(value_node, deep=deep)
            # next two lines differ from original
            if key in mapping:
                raise KeyError("Key \"{}\" was defined multiple times in config {}".
                               format(key, key_node.start_mark))
            mapping[key] = value
        return mapping


# pylint: disable-msg=abstract-method
class MpfLoader(Reader, Scanner, Parser, Composer, MpfConstructor, MpfResolver):

    """Config loader."""

    # pylint: disable-msg=super-init-not-called
    def __init__(self, stream, version=None, preserve_quotes=None):
        """Initialise loader."""
        del preserve_quotes
        del version
        Reader.__init__(self, stream, loader=self)
        Scanner.__init__(self, loader=self)
        Parser.__init__(self, loader=self)
        Composer.__init__(self, loader=self)
        MpfConstructor.__init__(self, loader=self)
        MpfResolver.__init__(self, loadumper=self)


class YamlInterface(FileInterface):

    """File interface for yaml files."""

    file_types = ['.yaml', '.yml']
    cache = False
    file_cache = dict()     # type: Dict[str, Any]

    __slots__ = []  # type: List[str]

    def load(self, filename, expected_version_str=None, halt_on_error=True) -> dict:
        """Load a YAML file from disk.

        Args:
        ----
            filename: The file to load.
            expected_version_str: Version string to expect or None for no check.
            halt_on_error: Boolean which controls what happens if the file
                can't be loaded. (Not found, invalid format, etc. If True, MPF
                will raise an error and exit. If False, an empty config
                dictionary will be returned.

        Returns a dictionary of the settings from this YAML file.
        """
        if self.cache and filename in self.file_cache:
            return copy.deepcopy(self.file_cache[filename])

        config = dict()     # type: dict

        try:
            self.log.debug("Loading file: %s", filename)

            with open(filename, encoding='utf8') as f:
                if expected_version_str:
                    version_str = f.readline().strip()
                    if version_str != expected_version_str:
                        raise AssertionError("Version mismatch. Expected: {} Actual: {} Files: {}".format(
                            expected_version_str, version_str, filename))
                config = self.process(f)
        except MarkedYAMLError as e:
            mark = e.problem_mark
            msg = "YAML error found in file {}. Line {}, " \
                  "Position {}: {}".format(filename, mark.line + 1 if mark else None,
                                           mark.column + 1 if mark else None, e)

            if halt_on_error:
                raise ValueError(msg)

            self.log.warning(msg)
        except Exception as e:   # pylint: disable-msg=broad-except
            msg = "Error found in file {}: {}".format(filename, e)

            if halt_on_error:
                raise ValueError(msg)
            self.log.warning(msg)

        if self.cache and config:
            self.file_cache[filename] = copy.deepcopy(config)

        return config

    @staticmethod
    def process(data_string: Iterable[str]) -> dict:
        """Parse yaml from a string."""
        return yaml.load(data_string, Loader=MpfLoader)

    def save(self, filename: str, data: dict) -> None:   # pragma: no cover
        """Save config to yaml file."""
        data_str = yaml.dump(data, default_flow_style=False)
        if not data_str:
            raise AssertionError("Failed to serialize data.")
        with open(filename, 'w', encoding='utf8') as output_file:
            output_file.write(data_str)



"""Base class used for things that "play" from the config files, such as WidgetPlayer, SlidePlayer, etc."""
import abc
import asyncio
from functools import partial
from typing import List

from mpf.core.machine import MachineController
from mpf.core.mode import Mode
from mpf.core.logging import LogMixin
from mpf.exceptions.config_file_error import ConfigFileError

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.placeholder_manager import BoolTemplate   # pylint: disable-msg=cyclic-import,unused-import
    from typing import Dict     # pylint: disable-msg=cyclic-import,unused-import


class ConfigPlayer(LogMixin, metaclass=abc.ABCMeta):

    """Base class for players which play things based on config."""

    config_file_section = None          # type: str
    show_section = None                 # type: str
    machine_collection_name = None      # type: str

    __slots__ = ["device_collection", "machine", "mode_event_keys", "instances"]

    def __init__(self, machine):
        """Initialise config player."""
        super().__init__()
        self.device_collection = None

        self.machine = machine      # type: MachineController

        # MPF only
        if hasattr(self.machine, "show_controller") and self.show_section:
            self.machine.show_controller.show_players[self.show_section] = self

        self._add_handlers()

        self.configure_logging(self.config_file_section)

        self.mode_event_keys = dict()
        self.instances = dict()
        self.instances['_global'] = dict()
        self.instances['_global'][self.config_file_section] = dict()

    def _add_handlers(self):
        self.machine.events.add_handler('init_phase_1', self._initialize_mode_handlers, priority=20)
        self.machine.events.add_handler('init_phase_1', self._initialise_system_wide, priority=1)

    def __repr__(self):
        """Return string representation."""
        return 'ConfigPlayer.{}/{}'.format(self.config_file_section, self.show_section)

    def _initialize_mode_handlers(self, **kwargs):
        del kwargs
        self.machine.mode_controller.register_load_method(
            self.process_mode_config, self.config_file_section)

        self.machine.mode_controller.register_start_method(
            self.mode_start, self.config_file_section)

        # Look through the machine config for config_player sections and
        # for shows to validate and process

        # if self.config_file_section in self.machine.config:

    def _initialise_system_wide(self, **kwargs):
        del kwargs
        if self.machine_collection_name:
            self.device_collection = getattr(self.machine,
                                             self.machine_collection_name)
        else:
            self.device_collection = None

        if (self.config_file_section in self.machine.config and
                self.machine.config[self.config_file_section]):

            # Validate
            self.machine.config[self.config_file_section] = (
                self.validate_config(
                    self.machine.config[self.config_file_section]))

            self.register_player_events(
                self.machine.config[self.config_file_section])

    def validate_config(self, config):
        """Validate this player's section of a config file (either a machine-wide config or a mode config).

        Args:
        ----
            config: A dict of the contents of this config_player's section
            from the config file. It's assumed that keys are event names, and
            values are settings for what this config_player does when that
            event is posted.

        Returns: A dict in the same format, but passed through the config
            validator.
        """
        # called first, before config file is cached. Not called if config file
        # is read from cache
        validated_config = dict()

        for event, settings in config.items():
            validated_config[event] = self.validate_config_entry(settings, event)

        return validated_config

    def validate_config_entry(self, settings, name):
        """Validate one entry of this player."""
        raise NotImplementedError("implement")

    def _parse_config(self, config, name):
        if config is None:
            raise AssertionError("Empty config player {}".format(name))
        if isinstance(config, (str, int, float)):
            # express config, convert to full
            config = self.get_express_config(config)
        elif isinstance(config, list):
            # list config. convert from list to full
            config = self.get_list_config(config)   # pylint: disable-msg=assignment-from-no-return

        if not isinstance(config, dict):
            raise AssertionError("Player config for {} is supposed to be dict. Config: {}".
                                 format(name, str(config)))

        return self.get_full_config(config)

    def process_mode_config(self, config, root_config_dict, mode, **kwargs):
        """Parse mode config."""
        del kwargs
        # handles validation and processing of mode config
        config = self.validate_config(config)
        root_config_dict[self.config_file_section] = config
        if mode.name not in self.instances:
            self.instances[mode.name] = dict()
        if self.config_file_section not in self.instances[mode.name]:
            self.instances[mode.name][self.config_file_section] = dict()

    def _get_full_context(self, context):
        return context + "." + self.config_file_section

    def _get_instance_dict(self, context):
        try:
            return self.instances[context][self.config_file_section]
        except KeyError:
            self.warning_log("Config player {} is missing context {}".format(self.config_file_section, context))
            return {}

    def _reset_instance_dict(self, context):
        if context not in self.instances:
            self.instances[context] = {}
        self.instances[context][self.config_file_section] = dict()

    @classmethod
    def process_config(cls, config, **kwargs):
        """Process system-wide config.

        Called every time mpf starts, regardless of whether config was built
        from cache or config files.
        """
        del kwargs
        return config

    def get_list_config(self, value) -> List[str]:
        """Parse config list."""
        del value
        raise AssertionError("Player {} does not support lists.".format(self.config_file_section))

    @abc.abstractmethod
    def get_express_config(self, value) -> dict:
        """Parse short config version.

        Implements "express" settings for this config_player which is what
        happens when a config is passed as a string instead of a full config
        dict. (This is detected automatically and this method is only called
        when the config is not a dict.)

        For example, the led_player uses the express config to parse a string
        like 'ff0000-f.5s' and translate it into:

        color: 220000
        fade: 500

        Since every config_player is different, this method raises a
        NotImplementedError and most be configured in the child class.

        Args:
        ----
            value: The single line string value from a config file.

        Returns a dictionary (which will then be passed through the config
        validator).
        """
        raise NotImplementedError(self.config_file_section)

    def get_full_config(self, value):
        """Return full config."""
        return self.machine.config_validator.validate_config(
            self.config_file_section, value, base_spec='config_player_common')

    def mode_start(self, config, priority, mode):
        """Add events for mode."""
        event_keys = self.register_player_events(config, mode, priority)

        self.mode_event_keys[mode] = event_keys

        return self.mode_stop, mode

    def mode_stop(self, mode):
        """Remove events for mode."""
        self.unload_player_events(self.mode_event_keys.pop(mode, list()))
        self.clear_context(mode.name)

    def clear_context(self, context):
        """Clear the context."""

    @staticmethod
    def is_entry_valid_outside_mode(settings) -> bool:
        """Return true if this entry may run without a game and player."""
        del settings
        return True

    # pylint: disable-msg=too-many-arguments
    def _create_subscription(self, template_str, subscription_list, settings, priority, mode):
        template = self.machine.placeholder_manager.build_raw_template(template_str)
        if mode:
            context = mode.name
            actual_priority = priority + mode.priority
        else:
            context = "_global"
            actual_priority = priority

        key = template_str

        self._update_subscription(template, subscription_list, settings, actual_priority, context, key, None)

    # pylint: disable-msg=too-many-arguments
    def _update_subscription(self, template, subscription_list, settings, priority, context, key, future):
        if future:
            try:
                future.result()
            except asyncio.CancelledError:
                return

        value, subscription = template.evaluate_and_subscribe([])
        subscription_list[template] = subscription
        subscription.add_done_callback(
            partial(self._update_subscription, template, subscription_list, settings, priority, context, key))
        self.handle_subscription_change(value, settings, priority, context, key)

    # pylint: disable-msg=no-self-use
    # pylint: disable-msg=too-many-arguments
    def handle_subscription_change(self, value, settings, priority, context, key):
        """Handle the change of a subscription."""
        del value
        del settings
        del priority
        del context
        del key
        raise AssertionError("Subscriptions are not supported in this player ({}).".format(self.__class__))

    def register_player_events(self, config, mode: Mode = None, priority=0):
        """Register events for standalone player."""
        # config is localized
        key_list = list()
        subscription_list = dict()      # type: Dict[BoolTemplate, asyncio.Future]

        if config:
            for event, settings in config.items():
                # prevent runtime crashes
                if (not mode or (mode and not mode.is_game_mode)) and not self.is_entry_valid_outside_mode(settings):
                    raise ConfigFileError("Section not valid outside of game modes. {} {}:{} Mode: {}".format(
                        self, event, settings, mode
                    ), 1, self.config_file_section)
                if event.startswith("{") and event.endswith("}"):
                    condition = event[1:-1]
                    self._create_subscription(condition, subscription_list, settings, priority, mode)
                else:
                    if mode and event in mode.config['mode']['start_events']:
                        self.machine.log.error(
                            "{0} mode's {1}: section contains a \"{2}:\" event "
                            "which is also in the start_events: for the {0} mode. "
                            "Change the {1}: {2}: event name to "
                            "\"mode_{0}_started:\"".format(
                                mode.name, self.config_file_section, event))

                        raise ValueError(
                            "{0} mode's {1}: section contains a \"{2}:\" event "
                            "which is also in the start_events: for the {0} mode. "
                            "Change the {1}: {2}: event name to "
                            "\"mode_{0}_started:\"".format(
                                mode.name, self.config_file_section, event))

                    key_list.append(
                        self.machine.events.add_handler(
                            event=event,
                            handler=self.config_play_callback,
                            calling_context=event,
                            priority=priority,
                            mode=mode,
                            settings=settings))

        return key_list, subscription_list

    def unload_player_events(self, key_list):
        """Remove event for standalone player."""
        for future in key_list[1].values():
            future.cancel()
        self.machine.events.remove_handlers_by_keys(key_list[0])

    def config_play_callback(self, settings, calling_context, priority=0, mode=None, **kwargs):
        """Handle play callback for standalone player."""
        # called when a config_player event is posted
        if mode:
            if not mode.active:
                # It's possible that an earlier event could have stopped the
                # mode before this event was handled, so just double-check to
                # make sure the mode is still active before proceeding.
                return None

            # calculate the base priority, which is a combination of the mode
            # priority and any priority value
            priority += mode.priority
            context = mode.name
        else:
            context = "_global"

        return self.play(settings=settings, context=context, calling_context=calling_context, priority=priority,
                         **kwargs)

    # pylint: disable-msg=too-many-arguments
    def show_play_callback(self, settings, priority, calling_context, show_tokens, context, start_time):
        """Handle show callback."""
        try:
            instance_dict = self.instances[context][self.config_file_section]
            del instance_dict
        except KeyError:
            # called from a show step
            if context not in self.instances:
                self.instances[context] = dict()

            if self.config_file_section not in self.instances[context]:
                self.instances[context][self.config_file_section] = dict()

        self.play(settings=settings, priority=priority, calling_context=calling_context,
                  show_tokens=show_tokens, context=context, start_time=start_time)

    def show_stop_callback(self, context):
        """Handle show stop."""
        self.clear_context(context)

    @abc.abstractmethod
    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Directly play player."""
        # **kwargs since this is an event callback
        raise NotImplementedError

    def check_delayed_play(self, key, settings, context, calling_context, priority=0, **kwargs):
        """Check the settings dict for a 'delay' property and re-post the play event later.

        This method will return True if it has scheduled the play for later, in which case the
        calling function should not proceed. Otherwise, the calling function should continue
        with its play behavior.
        """
        delay = settings.get("delay")
        if delay:
            del settings["delay"]
            self.machine.clock.schedule_once(
                lambda dt: self.play({key: settings}, context, calling_context, priority, **kwargs), delay)
            return True
        return False

    def expand_config_entry(self, settings):
        """Expend objects in config entry idempotently."""
        return settings



"""Maintains the ball count for a ball device."""
import asyncio
from typing import Optional

from mpf.devices.ball_device.physical_ball_counter import PhysicalBallCounter, EjectTracker
from mpf.devices.ball_device.entrance_switch_counter import EntranceSwitchCounter
from mpf.devices.ball_device.switch_counter import SwitchCounter

from mpf.core.utility_functions import Util
from mpf.devices.ball_device.ball_device_state_handler import BallDeviceStateHandler


class BallCountHandler(BallDeviceStateHandler):

    """Handles the ball count in the device."""

    __slots__ = ["_is_counting", "_count_valid", "_revalidate", "_eject_started", "_eject_ended", "_has_balls",
                 "_ball_count", "_ball_count_changed_futures", "counter"]

    def __init__(self, ball_device):
        """Initialise ball count handler."""
        super().__init__(ball_device)
        # inputs
        self._is_counting = asyncio.Lock()
        self._count_valid = asyncio.Event()
        self._revalidate = asyncio.Event()
        self._eject_started = asyncio.Event()
        self._eject_ended = asyncio.Event()
        self._has_balls = asyncio.Event()
        self._ball_count = 0
        self._ball_count_changed_futures = []
        self.counter = None  # type: Optional[PhysicalBallCounter]

    def wait_for_ball_count_changed(self):
        """Wait until ball count changed."""
        future = asyncio.Future()
        self._ball_count_changed_futures.append(future)
        return future

    def stop(self):
        """Stop counter."""
        super().stop()
        if self.counter:
            self.counter.stop()
            self.counter = None

    @property
    def handled_balls(self):
        """Return balls which are already handled."""
        return self._ball_count

    def _set_ball_count(self, count):
        self._ball_count = count
        # mirror variable at ball device for monitor
        self.ball_device.counted_balls = count
        if self._ball_count > 0:
            self._has_balls.set()
        else:
            self._has_balls.clear()

        self.machine.events.post("balldevice_{}_ball_count_changed".format(self.ball_device.name), balls=count)
        '''event: balldevice_(name)_ball_count_changed
        config_section: ball_devices
        class_label: ball_device

        desc: The ball count for device (name) just changed.

        This event may also be called without a change in some circumstances.

        args:

        balls: The number of new balls in this device.
        '''

        for future in self._ball_count_changed_futures:
            if not future.done():
                future.set_result(count)

        # reset futures
        self._ball_count_changed_futures = []

    async def initialise(self):
        """Initialise handler."""
        counter_config = self.ball_device.config.get("counter", {})
        if counter_config:
            counter_class = Util.string_to_class(counter_config["class"])
        elif self.ball_device.config.get('ball_switches'):
            counter_class = SwitchCounter
        else:
            counter_class = EntranceSwitchCounter

        self.counter = counter_class(self.ball_device, self.ball_device.config.get("counter", {}))

        self._ball_count = await self.counter.count_balls()
        # on start try to reorder balls if count is unstable
        if self.counter.is_count_unreliable():
            self.debug_log("BCH: Count is unstable. Trying to reorder balls.")
            await self.ball_device.ejector.reorder_balls()
            # recount
            self._ball_count = await self.counter.count_balls()

        if self._ball_count > 0:
            self._has_balls.set()
        self.ball_device.counted_balls = self._ball_count
        await super().initialise()
        self._count_valid.set()

    @property
    def has_ball(self) -> bool:
        """Return true if the device has at least one ball."""
        return self._ball_count > 0

    @property
    def is_full(self) -> bool:
        """Return true if the device is full."""
        if not self.counter:
            raise asyncio.CancelledError
        return self.counter.capacity - self._ball_count <= 0

    async def wait_for_ball(self):
        """Wait until the device has a ball."""
        if self.has_ball:
            self.debug_log("We have %s balls.", self._ball_count)
            return

        self.debug_log("No ball found. Waiting for balls.")

        # wait until we have more than 0 balls
        if not self.counter:
            raise asyncio.CancelledError
        ball_changes = asyncio.ensure_future(self.counter.wait_for_ball_count_changes(0))
        new_balls = await ball_changes

        # update count
        old_ball_count = self._ball_count
        self._ball_count = new_balls
        if new_balls > old_ball_count:
            self.debug_log("BCH: Found %s new balls", new_balls - old_ball_count)
            # handle new balls via incoming balls handler
            for _ in range(new_balls - old_ball_count):
                await self.ball_device.incoming_balls_handler.ball_arrived()
            self._set_ball_count(new_balls)

        self.debug_log("A ball arrived. Progressing.")

    # pylint: disable-msg=inconsistent-return-statements
    async def wait_for_ready_to_receive(self, source):
        """Wait until this device is ready to receive a ball."""
        while True:
            if not self.counter:
                raise asyncio.CancelledError
            free_space = self.counter.capacity - self._ball_count
            incoming_balls = self.ball_device.incoming_balls_handler.get_num_incoming_balls()
            if free_space <= incoming_balls:
                self.debug_log(
                    "Not ready to receive from %s. Not enough space. "
                    "Free space %s (Capacity: %s, Balls: %s), incoming_balls: %s",
                    source, free_space, self.counter.capacity, self._ball_count,
                    incoming_balls)
                await self.wait_for_ball_count_changed()
                continue

            if not self.counter:
                raise asyncio.CancelledError
            if not self.counter.is_ready_to_receive:
                self.debug_log(
                    "Not ready to receive from %s. Waiting on counter to become ready. "
                    "Free space %s (Capacity: %s, Balls: %s), incoming_balls: %s",
                    source, free_space, self.counter.capacity, self._ball_count,
                    incoming_balls)
                # wait for the counter to be ready
                await self.counter.wait_for_ready_to_receive()
                continue

            # wait until any eject conditions have passed which would break on an incoming ball
            if not self.ball_device.outgoing_balls_handler.is_ready_to_receive:
                self.debug_log(
                    "Not ready to receive from %s. Target is currently ejecting. "
                    "Free space %s (Capacity: %s, Balls: %s), incoming_balls: %s",
                    source, free_space, self.counter.capacity, self._ball_count,
                    incoming_balls)
                await self.ball_device.outgoing_balls_handler.wait_for_ready_to_receive()
                continue

            if not self.counter:
                raise asyncio.CancelledError
            self.debug_log("Ready to receive from %s. Free space %s (Capacity: %s, Balls: %s), incoming_balls: %s",
                           source, free_space, self.counter.capacity, self._ball_count,
                           incoming_balls)
            return True

    async def start_eject(self, already_left=False) -> EjectTracker:
        """Start eject."""
        await self.ball_device.incoming_balls_handler.start_eject()
        await self._is_counting.acquire()
        self._eject_started.set()
        self.debug_log("Entered eject mode.")

        eject_process = EjectTracker(self, already_left)
        if already_left:
            self._set_ball_count(self._ball_count + 1)
            await eject_process.will_eject()
        return eject_process

    async def end_eject(self, eject_process: EjectTracker, ball_left):
        """End eject."""
        eject_process.cancel()
        self.debug_log("Exited eject mode. Eject success: %s", ball_left)
        if ball_left:
            self._set_ball_count(self._ball_count - 1)
        self._eject_started.clear()
        self._is_counting.release()
        self.ball_device.incoming_balls_handler.end_eject()

    async def _run(self):
        if not self.counter:
            raise asyncio.CancelledError
        changes = self.counter.register_change_stream()
        while True:
            # wait for ball changes
            ball_changes = asyncio.ensure_future(changes.get())
            revalidate_future = asyncio.ensure_future(self._revalidate.wait())
            await Util.first([ball_changes, revalidate_future, self._eject_started.wait()])
            self._revalidate.clear()

            # get lock and update count
            await self._is_counting.acquire()

            if not self.counter:
                raise asyncio.CancelledError
            new_balls = await self.counter.count_balls()

            # try to re-order the device if count is unstable
            if not self.counter:
                raise asyncio.CancelledError
            if self.counter.is_count_unreliable():
                self.debug_log("BCH: Count is unstable. Trying to reorder balls.")
                await self.ball_device.ejector.reorder_balls()
                new_balls = await self.counter.count_balls()

            self.debug_log("BCH: Counting. New count: %s Old count: %s", new_balls, self._ball_count)

            # when jammed do not trust other switches except the jam. keep old count
            if not self.counter:
                raise asyncio.CancelledError
            if not self.counter.is_count_unreliable():
                # otherwise handle balls
                old_ball_count = self._ball_count
                if new_balls > old_ball_count:
                    self.debug_log("BCH: Found %s new balls", new_balls - old_ball_count)
                    self._set_ball_count(new_balls)
                    # handle new balls via incoming balls handler
                    for _ in range(new_balls - old_ball_count):
                        await self.ball_device.incoming_balls_handler.ball_arrived()
                elif new_balls < old_ball_count:
                    await self._handle_missing_balls(new_balls, old_ball_count - new_balls)

            self._is_counting.release()
            self._count_valid.set()

    async def _handle_missing_balls(self, new_balls, missing_balls):
        if self.ball_device.outgoing_balls_handler.is_idle:
            if self.ball_device.config['mechanical_eject']:
                self.debug_log("BCH: Lost %s balls. Assuming mechanical eject.", missing_balls)
                self._set_ball_count(new_balls)
                await self.ball_device.handle_mechanial_eject_during_idle()
            else:
                try:
                    if not self.counter:
                        raise asyncio.CancelledError
                    await asyncio.wait_for(self.counter.wait_for_ball_activity(),
                                           timeout=self.ball_device.config['idle_missing_ball_timeout'])
                except asyncio.TimeoutError:
                    self.debug_log("BCH: Lost %s balls", missing_balls)
                    self._set_ball_count(new_balls)
                    for _ in range(missing_balls):
                        await self.ball_device.lost_idle_ball()
                else:
                    self._revalidate.set()
        else:
            self.debug_log("Lost ball %s balls between ejects. Ignoring.", missing_balls)

    async def wait_for_count_is_valid(self):
        """Wait until count is valid."""
        self._count_valid.clear()
        self._revalidate.set()
        # wait for ball_counter to become ready
        await self._count_valid.wait()

    async def entrance_during_eject(self):
        """Received an entrance during eject."""
        await self.ball_device.incoming_balls_handler.ball_arrived()
        self._set_ball_count(self._ball_count + 1)



from unittest.mock import MagicMock

import sys
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class MockEvent:

    """Copy of the irc Events class."""

    def __init__(self, type, source, target, arguments=None, tags=None):

        self.type = type
        self.source = source
        self.target = target
        if arguments is None:
            arguments = []
        self.arguments = arguments
        if tags is None:
            tags = []
        self.tags = tags


class MockSingleServerIRCBot():

    """Mock server."""

    def __init__(self, server_list, nickname, realname, reconnection_interval=None, recon=None, **connect_params):
        self.connection = MagicMock()

    def start(self):
        pass


class TestTwitchClient(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/twitch_client/'

    def setUp(self):
        self.machine_config_patches['mpf']['plugins'] = ['mpf.plugins.twitch_bot.TwitchBot']
        sys.modules['irc.bot'] = MagicMock()
        sys.modules['irc.bot'].SingleServerIRCBot = MockSingleServerIRCBot
        super().setUp()

    def tearDown(self):
        del sys.modules['irc.bot']

    def test_twitch_chat(self):
        """Test connect and event posting."""
        self.mock_event("twitch_chat_message")
        tags = [
            {"key": "msg-id", "value": None},
            {"key": "bits", "value": None},
            {"key": "display-name", "value": "Some User"},
            {"key": "msg-param-months", "value": None},
            {"key": "message", "value": None},
        ]
        event = MockEvent("pubmsg", "some_user", "bot", ["Hello Bot"], tags)
        self.machine.plugins[0].client.on_pubmsg("some_channel", event)
        self.advance_time_and_run(.1)
        self.assertEventCalled("twitch_chat_message")
        self.assertMachineVarEqual("Some User", "twitch_last_chat_user")
        self.assertMachineVarEqual("Hello Bot", "twitch_last_chat_message")

    def test_twitch_bits(self):
        """Test connect and event posting."""
        self.mock_event("twitch_bit_donation")
        tags = [
            {"key": "msg-id", "value": None},
            {"key": "bits", "value": 42},
            {"key": "display-name", "value": "Some User"},
            {"key": "msg-param-months", "value": None},
            {"key": "message", "value": None},
        ]
        event = MockEvent("pubmsg", "some_user", "bot", ["Hello Bot"], tags)
        self.machine.plugins[0].client.on_pubmsg("some_channel", event)
        self.advance_time_and_run(.1)
        self.assertEventCalled("twitch_bit_donation")
        self.assertMachineVarEqual("Some User", "twitch_last_bits_user")
        self.assertMachineVarEqual(42, "twitch_last_bits_amount")

    def test_twitch_sub(self):
        """Test connect and event posting."""
        self.mock_event("twitch_subscription")
        tags = [
            {"key": "msg-id", "value": "sub"},
            {"key": "bits", "value": None},
            {"key": "display-name", "value": "Some User"},
            {"key": "msg-param-months", "value": 1},
            {"key": "message", "value": "Sub message"},
        ]
        event = MockEvent("usernotice", "some_user", "bot", ["Hello Bot"], tags)
        self.machine.plugins[0].client.on_usernotice("some_channel", event)
        self.advance_time_and_run(.1)
        self.assertEventCalled("twitch_subscription")

    def test_twitch_resub(self):
        """Test connect and event posting."""
        self.mock_event("twitch_subscription")
        tags = [
            {"key": "msg-id", "value": "resub"},
            {"key": "bits", "value": None},
            {"key": "display-name", "value": "Some User"},
            {"key": "msg-param-months", "value": 42},
            {"key": "message", "value": "Sub message"},
        ]
        event = MockEvent("usernotice", "some_user", "bot", ["Hello Bot"], tags)
        self.machine.plugins[0].client.on_usernotice("some_channel", event)
        self.advance_time_and_run(.1)
        self.assertEventCalled("twitch_subscription")

    def test_twitch_gift_sub(self):
        """Test connect and event posting."""
        self.mock_event("twitch_subscription")
        tags = [
            {"key": "msg-id", "value": "subgift"},
            {"key": "bits", "value": None},
            {"key": "display-name", "value": "Some User"},
            {"key": "msg-param-months", "value": 1},
            {"key": "message", "value": "Sub message"},
        ]
        event = MockEvent("usernotice", "some_user", "bot", ["Hello Bot"], tags)
        self.machine.plugins[0].client.on_usernotice("some_channel", event)
        self.advance_time_and_run(.1)
        self.assertEventCalled("twitch_subscription")

    def test_twitch_anon_gift_sub(self):
        """Test connect and event posting."""
        self.mock_event("twitch_subscription")
        tags = [
            {"key": "msg-id", "value": "anonsubgift"},
            {"key": "bits", "value": None},
            {"key": "display-name", "value": "Some User"},
            {"key": "msg-param-months", "value": 42},
            {"key": "message", "value": "Sub message"},
        ]
        event = MockEvent("usernotice", "some_user", "bot", ["Hello Bot"], tags)
        self.machine.plugins[0].client.on_usernotice("some_channel", event)
        self.advance_time_and_run(.1)
        self.assertEventCalled("twitch_subscription")

    def test_twitch_sub_no_message(self):
        """Test connect and event posting."""
        self.mock_event("twitch_subscription")
        tags = [
            {"key": "msg-id", "value": "sub"},
            {"key": "bits", "value": None},
            {"key": "display-name", "value": "Some User"},
            {"key": "msg-param-months", "value": 1},
            {"key": "message", "value": "Sub message"},
        ]
        event = MockEvent("usernotice", "some_user", "bot", [], tags)
        self.machine.plugins[0].client.on_usernotice("some_channel", event)
        self.advance_time_and_run(.1)
        self.assertEventCalled("twitch_subscription")

    def test_twitch_raid(self):
        """Test connect and event posting."""
        self.mock_event("twitch_raid")
        tags = [
            {"key": "msg-id", "value": "raid"},
            {"key": "bits", "value": None},
            {"key": "display-name", "value": "Some User"},
            {"key": "msg-param-months", "value": 1},
            {"key": "message", "value": "Sub message"},
        ]
        event = MockEvent("usernotice", "some_user", "bot", ["Hello Bot"], tags)
        self.machine.plugins[0].client.on_usernotice("some_channel", event)
        self.advance_time_and_run(.1)
        self.assertEventCalled("twitch_raid")



from mpf.tests.MpfTestCase import MpfTestCase


class TestPlayfieldTransfer(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/playfield_transfer/'

    def testBallPassThroughSwitch(self):
        # test pass from pf1 to pf2
        pf1 = self.machine.ball_devices['playfield1']
        pf2 = self.machine.ball_devices['playfield2']

        self.set_num_balls_known(4)
        pf1.balls = 2
        pf1.available_balls = 2
        pf2.balls = 2
        pf2.available_balls = 2

        self.machine.switch_controller.process_switch("s_transfer", 1)
        self.advance_time_and_run(2)

        self.assertEqual(1, pf1.balls)
        self.assertEqual(1, pf1.available_balls)
        self.assertEqual(3, pf2.balls)
        self.assertEqual(3, pf2.available_balls)

    def testBallPassThroughEvent(self):
        # test pass from pf1 to pf2
        pf1 = self.machine.ball_devices['playfield1']
        pf2 = self.machine.ball_devices['playfield2']

        self.set_num_balls_known(4)
        pf1.balls = 2
        pf1.available_balls = 2
        pf2.balls = 2
        pf2.available_balls = 2

        self.post_event("transfer_ball")
        self.advance_time_and_run(2)

        self.assertEqual(1, pf1.balls)
        self.assertEqual(1, pf1.available_balls)
        self.assertEqual(3, pf2.balls)
        self.assertEqual(3, pf2.available_balls)



"""Holds various Version strings of MPF.

This modules holds the MPF version strings, including the version of BCP it
needs and the config file version it needs.

It's used internally for all sorts of things, from printing the output of the
`mpf --version` command, to making sure any processes connected via BCP are
the proper versions, to automatically triggering new builds and deployments to
PyPI.

"""

__version__ = '0.55.0-dev.39'
'''The full version of MPF.'''

__short_version__ = '0.55'
'''The major.minor version of MPF.'''

__bcp_version__ = '1.1'
'''The version of BCP this build of MPF uses.'''

__config_version__ = '5'
'''The config file version this build of MPF uses.'''

__show_version__ = '5'
'''The show format version this build of MPF uses.'''

# pylint: disable-msg=invalid-name
version = "MPF v{}".format(__version__)
'''A friendly version string for this build of MPF.'''

# pylint: disable-msg=invalid-name
extended_version = "MPF v{}, Config version:{}, Show version: {}, " \
                   "BCP version:{}".format(__version__, __config_version__,
                                           __show_version__, __bcp_version__)
'''An extended version string that includes the MPF version, show version,
and BCP versions used in this build of MPF.'''

if "dev" in __version__:
    # pylint: disable-msg=invalid-name
    log_url = "https://docs.missionpinball.org/en/dev/logs/{}.html"
else:
    # pylint: disable-msg=invalid-name
    log_url = "https://docs.missionpinball.org/en/{}/logs/{{}}.html".format(__short_version__)

__api__ = ['version',
           '__short_version__',
           '__bcp_version__',
           '__config_version__',
           '__show_version__',
           'version',
           'extended_version',
           'log_url']



"""Event Config Player."""
from copy import deepcopy

from mpf.core.placeholder_manager import TextTemplate

from mpf.config_players.flat_config_player import FlatConfigPlayer
from mpf.core.delays import DelayManager
from mpf.core.utility_functions import Util


class EventPlayer(FlatConfigPlayer):

    """Posts events based on config."""

    config_file_section = 'event_player'
    show_section = 'events'

    __slots__ = ["delay"]

    def __init__(self, machine):
        """Initialise EventPlayer."""
        super().__init__(machine)
        self.delay = DelayManager(self.machine)

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Post (delayed) events."""
        for event, s in settings.items():
            s = deepcopy(s)
            event_dict = self.machine.placeholder_manager.parse_conditional_template(event)

            if event_dict.condition and not event_dict.condition.evaluate(kwargs):
                continue

            if event_dict.number:
                delay = Util.string_to_ms(event_dict.number)
                self.delay.add(callback=self._post_event, ms=delay,
                               event=event_dict.name, s=s, **kwargs)
            else:
                self._post_event(event_dict.name, s, **kwargs)

    def _post_event(self, event, s, **kwargs):
        event_name_placeholder = TextTemplate(self.machine, event.replace("(", "{").replace(")", "}"))
        for key, param in s.items():
            if isinstance(param, dict):
                s[key] = self._evaluate_event_param(param, kwargs)
        self.machine.events.post(event_name_placeholder.evaluate(kwargs), **s)

    def _evaluate_event_param(self, param, kwargs):
        if param.get("type") == "float":
            placeholder = self.machine.placeholder_manager.build_float_template(param["value"])
        elif param.get("type") == "int":
            placeholder = self.machine.placeholder_manager.build_int_template(param["value"])
        elif param.get("type") == "bool":
            placeholder = self.machine.placeholder_manager.build_bool_template(param["value"])
        else:
            placeholder = self.machine.placeholder_manager.build_string_template(param["value"])
        return placeholder.evaluate(kwargs)

    def get_list_config(self, value):
        """Parse list."""
        result = {}
        for event in value:
            result[event] = {}
        return result

    def get_express_config(self, value):
        """Parse short config."""
        return self.get_list_config(Util.string_to_event_list(value))



"""Pinproc object for tests."""

class FakePinProcModule:

    DriverCount = 256

    EventTypeAccelerometerIRQ = 11
    EventTypeAccelerometerX = 8
    EventTypeAccelerometerY = 9
    EventTypeAccelerometerZ = 10
    EventTypeBurstSwitchClosed = 7
    EventTypeBurstSwitchOpen = 6
    EventTypeDMDFrameDisplayed = 5
    EventTypeSwitchClosedDebounced = 1
    EventTypeSwitchClosedNondebounced = 3
    EventTypeSwitchOpenDebounced = 2
    EventTypeSwitchOpenNondebounced = 4

    MachineTypeCustom = 1
    MachineTypeInvalid = 0
    MachineTypePDB = 7
    MachineTypeSternSAM = 6
    MachineTypeSternWhitestar = 5
    MachineTypeWPC = 3
    MachineTypeWPC95 = 4
    MachineTypeWPCAlphanumeric = 2

    SwitchCount = 255
    SwitchNeverDebounceFirst = 192
    SwitchNeverDebounceLast = 255

    def __init__(self):
        self.pinproc = FakePinProc()

    def driver_state_pulse(self, driver, milliseconds):
        driver["state"] = 1
        driver["timeslots"] = 0
        driver["waitForFirstTimeSlot"] = False
        driver["outputDriveTime"] = milliseconds
        driver["patterOnTime"] = 0
        driver["patterOffTime"] = 0
        driver["patterEnable"] = False
        driver["futureEnable"] = False
        return driver

    def driver_state_disable(self, driver):
        driver["state"] = 0
        driver["timeslots"] = 0
        driver["waitForFirstTimeSlot"] = False
        driver["outputDriveTime"] = 0
        driver["patterOnTime"] = 0
        driver["patterOffTime"] = 0
        driver["patterEnable"] = False
        driver["futureEnable"] = False
        return driver

    def driver_state_patter(self, driver, millisecondsOn, millisecondsOff, originalOnTime, now):
        driver["state"] = True
        driver["timeslots"] = 0
        driver["waitForFirstTimeSlot"] = not now
        driver["outputDriveTime"] = originalOnTime
        driver["patterOnTime"] = millisecondsOn
        driver["patterOffTime"] = millisecondsOff
        driver["patterEnable"] = True
        driver["futureEnable"] = False
        return driver

    def driver_pulsed_patter(self, driver, millisecondsOn, millisecondsOff, milliseconds_overall_patter_time, now):
        driver["state"] = True
        driver["timeslots"] = 0
        driver["waitForFirstTimeSlot"] = not now
        driver["outputDriveTime"] = milliseconds_overall_patter_time
        driver["patterOnTime"] = millisecondsOn
        driver["patterOffTime"] = millisecondsOff
        driver["patterEnable"] = True
        driver["futureEnable"] = False
        return driver

    def normalize_machine_type(self, type):
        return 7

    def PinPROC(self, machine_type):
        return self.pinproc


class FakePinProc:

    """Behaves like pypinproc."""

    def __init__(self):
        self._memory = {
            0x00: {         # manager
                0x00: 0,            # chip id
                0x01: 0x00020006,   # version
                0x03: 0x01FF,       # dip switches
            },
            0x02: {         # switch controller
                0x1000: 0xA3,       # SW-16 Address 0 Reg 0
                0x1001: 0x00,       # SW-16 Address 0 Reg 1
                0x1040: 0xA3,       # SW-16 Address 1 Reg 0
                0x1041: 0x13,       # SW-16 Address 1 Reg 1
                0x1080: 0xA4,       # SW-16 Address 2 Reg 0
                0x1081: 0x00,       # SW-16 Address 2 Reg 1
            }
        }
        self._switches = [0, 1] + [0] * 100
        self._events = []

    def read_data(self, module, address):
        if module not in self._memory or address not in self._memory[module]:
            return 0
        return self._memory[module][address]

    def write_data(self, module, address, data):
        if module not in self._memory:
            self._memory[module] = {}
        self._memory[module][address] = data

    def switch_get_states(self):
        return self._switches

    def flush(self):
        pass

    def switch_update_rule(self, *args):
        return True

    def driver_update_group_config(self, *args):
        return True

    def driver_update_global_config(self, *args):
        return True

    def driver_update_state(self, *args):
        return True

    def driver_pulse(self, *args):
        return True

    def driver_schedule(self, *args):
        return True

    def driver_patter(self, *args):
        return True

    def driver_disable(self, *args):
        return True

    def reset(self, *args):
        self._events = []
        return True

    def watchdog_tickle(self):
        pass

    def close(self):
        pass

    def get_events(self):
        events = self._events
        self._events = []
        return events



"""A kickback device which will fire a ball back into the playfield."""
from typing import List

from mpf.core.device_monitor import DeviceMonitor
from mpf.devices.autofire import AutofireCoil


@DeviceMonitor(_enabled="enabled")
class Kickback(AutofireCoil):

    """A kickback device which will fire a ball back into the playfield."""

    config_section = 'kickbacks'
    collection = 'kickbacks'
    class_label = 'kickback'

    __slots__ = []  # type: List[str]

    def _hit(self):
        """Post fired event."""
        super()._hit()
        if self._enabled:
            self.machine.events.post("kickback_{}_fired".format(self.name))
            '''event: kickback_(name)_fired

            desc: Kickback fired a ball.
            '''



from mpf.tests.MpfGameTestCase import MpfGameTestCase


class TestExtraBall(MpfGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/extra_ball/'

    def get_platform(self):
        return 'smart_virtual'

    def test_extra_ball(self):
        # tests basic EB functionality and also max_per_game setting
        self.mock_event('extra_ball_awarded')
        self.mock_event('extra_ball_eb1_awarded')
        self.mock_event('extra_ball_award_disabled')
        self.mock_event('extra_ball_eb1_award_disabled')
        self.mock_event("first_ball")

        self.fill_troughs()
        self.start_game()

        self.assertEventCalled("first_ball")
        self.mock_event("first_ball")

        # start mode
        self.post_event("start_mode1")

        # mode loaded. eb1 should be enabled
        self.assertTrue(self.machine.extra_balls["eb1"])
        self.assertTrue(self.machine.extra_balls["eb1"].player)
        self.assertTrue(self.machine.extra_balls["eb1"].enabled)
        self.assertEqual(1, self.machine.game.player.number)
        self.assertEqual(0, self.machine.game.player.extra_ball_eb1_awarded)
        self.assertEqual(0, self.machine.game.player.extra_balls)

        # stop mode
        self.post_event("stop_mode1")

        # nothing should happen
        self.post_event("award_eb1")
        self.assertEqual(1, self.machine.game.player.number)
        self.assertEqual(0, self.machine.game.player.extra_ball_eb1_awarded)
        self.assertEqual(0, self.machine.game.player.extra_balls)
        self.assertFalse(self.machine.extra_balls["eb1"].player)

        # start mode (again)
        self.post_event("start_mode1")

        self.assertTrue(self.machine.extra_balls["eb1"])
        self.assertTrue(self.machine.extra_balls["eb1"].player)
        self.assertTrue(self.machine.extra_balls["eb1"].enabled)
        self.assertEqual(1, self.machine.game.player.number)
        self.assertEqual(0, self.machine.game.player.extra_ball_eb1_awarded)
        self.assertEqual(0, self.machine.game.player.extra_balls)

        # player gets extra_ball
        self.post_event("award_eb1")
        self.assertEqual(1, self.machine.game.player.number)
        self.assertEqual(1, self.machine.game.player.ball)
        self.assertEqual(1, self.machine.game.player.extra_ball_eb1_num_awarded)
        self.assertEqual(1, self.machine.game.player.extra_balls)
        self.assertEventCalled('extra_ball_awarded')
        self.assertEventCalled('extra_ball_eb1_awarded')
        self.assertEventNotCalled('extra_ball_award_disabled')
        self.assertEventNotCalled('extra_ball_eb1_award_disabled')

        # but only once
        self.assertFalse(self.machine.extra_balls["eb1"].enabled)
        self.mock_event('extra_ball_awarded')
        self.mock_event('extra_ball_eb1_awarded')
        self.mock_event('extra_ball_award_disabled')
        self.mock_event('extra_ball_eb1_award_disabled')
        self.post_event("award_eb1")
        self.assertEqual(1, self.machine.game.player.number)
        self.assertEqual(1, self.machine.game.player.ball)
        self.assertEqual(1, self.machine.game.player.extra_ball_eb1_num_awarded)
        self.assertEqual(1, self.machine.game.player.extra_balls)
        self.assertEventNotCalled('extra_ball_awarded')
        self.assertEventNotCalled('extra_ball_eb1_awarded')
        self.assertEventCalled('extra_ball_award_disabled')
        self.assertEventCalled('extra_ball_eb1_award_disabled')

        # drain and start new ball, but should still be ball 1 since the player
        # has an EB
        self.drain_all_balls()
        self.advance_time_and_run()
        self.assertEqual(1, self.machine.game.player.number)
        self.assertEventNotCalled("first_ball")

        # EB disabled events should be posted
        self.mock_event('extra_ball_awarded')
        self.mock_event('extra_ball_eb1_awarded')
        self.mock_event('extra_ball_award_disabled')
        self.mock_event('extra_ball_eb1_award_disabled')
        self.post_event('start_mode1')
        self.post_event('award_eb1')
        self.assertEqual(1, self.machine.game.player.number)
        self.assertEqual(1, self.machine.game.player.ball)
        self.assertEqual(1, self.machine.game.player.extra_ball_eb1_num_awarded)
        self.assertEqual(0, self.machine.game.player.extra_balls)
        self.assertEventNotCalled('extra_ball_awarded')
        self.assertEventNotCalled('extra_ball_eb1_awarded')
        self.assertEventCalled('extra_ball_award_disabled')
        self.assertEventCalled('extra_ball_eb1_award_disabled')

    def test_extra_ball_disabled(self):
        self.mock_event('extra_ball_awarded')
        self.mock_event('extra_ball_eb3_awarded')
        self.mock_event('extra_ball_award_disabled')
        self.mock_event('extra_ball_eb3_award_disabled')

        self.fill_troughs()
        self.start_game()
        self.post_event("start_mode1")

        # mode loaded. eb3 should not be enabled
        self.assertTrue(self.machine.extra_balls["eb3"])
        self.assertTrue(self.machine.extra_balls["eb3"].player)
        self.assertFalse(self.machine.extra_balls["eb3"].enabled)
        self.assertEqual(1, self.machine.game.player.number)
        self.assertEqual(0, self.machine.game.player.extra_ball_eb2_awarded)
        self.assertEqual(0, self.machine.game.player.extra_balls)

        # player gets extra_ball, but it's disabled
        self.post_event("award_eb3")
        self.assertEqual(1, self.machine.game.player.number)
        self.assertEqual(1, self.machine.game.player.ball)
        self.assertEqual(0, self.machine.game.player.extra_ball_eb3_num_awarded)
        self.assertEqual(0, self.machine.game.player.extra_balls)
        self.assertEventNotCalled('extra_ball_awarded')
        self.assertEventNotCalled('extra_ball_eb3_awarded')
        self.assertEventCalled('extra_ball_award_disabled')
        self.assertEventCalled('extra_ball_eb3_award_disabled')

    def test_extra_ball_lit_with_no_group(self):
        self.mock_event('extra_ball_eb2_lit')
        self.mock_event('extra_ball_eb2_awarded')

        self.fill_troughs()
        self.start_game()
        self.post_event('start_mode1')
        self.post_event('light_eb2')
        self.assertEventCalled('extra_ball_eb2_lit')
        self.assertEventNotCalled('extra_ball_eb2_awarded')

    def test_extra_ball_lit_with_no_group_disabled(self):
        self.mock_event('extra_ball_eb4_lit')
        self.mock_event('extra_ball_eb4_awarded')
        self.mock_event('extra_ball_award_disabled')
        self.mock_event('extra_ball_eb4_award_disabled')

        self.fill_troughs()
        self.start_game()
        self.post_event('start_mode1')
        self.post_event('light_eb4')
        self.assertEventNotCalled('extra_ball_eb4_lit')
        self.assertEventNotCalled('extra_ball_eb4_awarded')
        self.assertEventCalled('extra_ball_award_disabled')
        self.assertEventCalled('extra_ball_eb4_award_disabled')

    def test_group(self):
        # tests basic group functionality
        self.fill_troughs()
        self.start_two_player_game()
        self.post_event('start_mode1')

        self.assertEqual(self.machine.extra_balls["eb5"].group,
                         self.machine.extra_ball_groups["main"])
        self.assertEqual(self.machine.extra_balls["eb6"].group,
                         self.machine.extra_ball_groups["main"])

        # award a group EB, even though none are lit
        self.mock_event('extra_ball_group_main_awarded')

        self.post_event('award_group_eb')
        self.assertEqual(1, self.machine.game.player.extra_balls)
        self.assertEqual(0, self.machine.game.player.extra_ball_group_main_num_lit)
        self.assertEqual(1, self.machine.game.player.extra_ball_group_main_num_awarded_game)
        self.assertEqual(1, self.machine.game.player.extra_ball_group_main_num_awarded_ball)
        self.assertEventCalled('extra_ball_group_main_awarded')

        # light single EB, make sure it works and the group vars update
        self.mock_event('extra_ball_group_main_lit')
        self.mock_event('extra_ball_group_main_lit_awarded')
        self.mock_event('extra_ball_group_main_awarded')

        self.post_event('light_eb5')
        self.assertEventCalled('extra_ball_group_main_lit')
        self.assertEventCalled('extra_ball_group_main_lit_awarded')
        self.assertEqual(1, self.machine.game.player.extra_balls)
        self.assertEqual(1, self.machine.game.player.extra_ball_group_main_num_lit)
        self.assertEqual(1, self.machine.game.player.extra_ball_group_main_num_awarded_game)
        self.assertEqual(1, self.machine.game.player.extra_ball_group_main_num_awarded_ball)

        # light another EB, but max lit for the group is 1, so it should not
        # light, and we should get the disabled event
        self.mock_event('extra_ball_group_main_lit')
        self.mock_event('extra_ball_group_main_lit_awarded')
        self.mock_event('extra_ball_group_main_awarded')
        self.mock_event('extra_ball_group_main_award_disabled')

        self.post_event('light_eb6')
        self.assertEventNotCalled('extra_ball_group_main_lit')
        self.assertEventNotCalled('extra_ball_group_main_lit_awarded')
        self.assertEventNotCalled('extra_ball_group_main_awarded')
        self.assertEventCalled('extra_ball_group_main_award_disabled')
        self.assertEqual(1, self.machine.game.player.extra_balls)
        self.assertEqual(1,
            self.machine.game.player.extra_ball_group_main_num_lit)
        self.assertEqual(1,
             self.machine.game.player.extra_ball_group_main_num_awarded_game)
        self.assertEqual(1,
             self.machine.game.player.extra_ball_group_main_num_awarded_ball)

        # drain the ball, player has EB so it should still be their turn
        # EB group should be lit from before
        self.mock_event('extra_ball_group_main_lit')
        self.mock_event('extra_ball_group_main_lit_awarded')

        self.drain_all_balls()
        self.advance_time_and_run()
        self.assertEqual(1, self.machine.game.player.number)
        self.assertEventCalled('extra_ball_group_main_lit')
        self.assertEventNotCalled('extra_ball_group_main_lit_awarded')
        self.assertEqual(0, self.machine.game.player.extra_balls)
        self.assertEqual(1,
            self.machine.game.player.extra_ball_group_main_num_lit)
        self.assertEqual(1,
             self.machine.game.player.extra_ball_group_main_num_awarded_game)
        self.assertEqual(1,
             self.machine.game.player.extra_ball_group_main_num_awarded_ball)

        # award another EB. This maxes out the extra balls they can get this
        # ball, so the lit one should unlight
        self.mock_event('extra_ball_group_main_awarded')
        self.mock_event('extra_ball_group_main_lit')
        self.mock_event('extra_ball_group_main_lit_awarded')
        self.mock_event('extra_ball_group_main_unlit')

        self.post_event('award_group_eb')
        self.assertEventNotCalled('extra_ball_group_main_lit')
        self.assertEventNotCalled('extra_ball_group_main_lit_awarded')
        self.assertEventCalled('extra_ball_group_main_awarded')
        self.assertEventCalled('extra_ball_group_main_unlit')
        self.assertEqual(1, self.machine.game.player.extra_balls)
        self.assertEqual(1,
            self.machine.game.player.extra_ball_group_main_num_lit)
        self.assertEqual(2,
             self.machine.game.player.extra_ball_group_main_num_awarded_game)
        self.assertEqual(2,
             self.machine.game.player.extra_ball_group_main_num_awarded_ball)

        # drain, should still be player 1 and the EB should still be lit
        self.mock_event('extra_ball_group_main_lit')
        self.mock_event('extra_ball_group_main_lit_awarded')

        self.drain_all_balls()
        self.advance_time_and_run()
        self.assertEventCalled('extra_ball_group_main_lit')
        self.assertEventNotCalled('extra_ball_group_main_lit_awarded')
        self.assertEqual(0, self.machine.game.player.extra_balls)
        self.assertEqual(1,
            self.machine.game.player.extra_ball_group_main_num_lit)
        self.assertEqual(2,
             self.machine.game.player.extra_ball_group_main_num_awarded_game)
        self.assertEqual(2,
             self.machine.game.player.extra_ball_group_main_num_awarded_ball)

        # light another EB, but the max is 2 per ball, and even though this is
        # the next ball, it was an extra ball, so still Ball 1, so it should
        # not light
        self.mock_event('extra_ball_group_main_lit')
        self.mock_event('extra_ball_group_main_lit_awarded')
        self.mock_event('extra_ball_group_main_award_disabled')

        self.post_event('start_mode1')
        self.post_event('light_eb7')
        self.assertEventNotCalled('extra_ball_group_main_lit')
        self.assertEventNotCalled('extra_ball_group_main_lit_awarded')
        self.assertEventCalled('extra_ball_group_main_award_disabled')
        self.assertEqual(0, self.machine.game.player.extra_balls)
        self.assertEqual(1,
            self.machine.game.player.extra_ball_group_main_num_lit)
        self.assertEqual(2,
             self.machine.game.player.extra_ball_group_main_num_awarded_game)
        self.assertEqual(2,
             self.machine.game.player.extra_ball_group_main_num_awarded_ball)

        # drain and make sure we get to player 2
        self.drain_all_balls()
        self.advance_time_and_run()
        self.assertEqual(2, self.machine.game.player.number)

    def test_group_disabled_with_eb_enabled(self):
        self.fill_troughs()
        self.start_game()
        self.post_event('start_mode1')

        self.mock_event('extra_ball_group_disabled_eb_lit')
        self.mock_event('extra_ball_group_disabled_eb_lit_awarded')
        self.mock_event('extra_ball_group_disabled_eb_award_disabled')

        self.post_event('light_eb8')
        self.assertEventNotCalled('extra_ball_group_disabled_eb_lit')
        self.assertEventNotCalled('extra_ball_group_disabled_eb_lit_awarded')
        self.assertEventCalled('extra_ball_group_disabled_eb_award_disabled')

        self.mock_event('extra_ball_group_disabled_eb_award_disabled')
        self.mock_event('extra_ball_group_disabled_eb_awarded')
        self.mock_event('extra_ball_eb8_award_disabled')
        self.mock_event('extra_ball_eb8_awarded')

        self.post_event('award_eb8')
        self.assertEventCalled('extra_ball_group_disabled_eb_award_disabled')
        self.assertEventNotCalled('extra_ball_group_disabled_eb_awarded')
        self.assertEventCalled('extra_ball_eb8_award_disabled')
        self.assertEventNotCalled('extra_ball_eb8_awarded')

        self.assertEqual(0, self.machine.game.player.extra_balls)

    def test_group_max_per_game(self):
        self.fill_troughs()
        self.start_game()
        self.post_event('start_mode1')
        self.post_event('award_eb9')
        self.post_event('award_eb9')
        self.post_event('award_eb9')
        self.assertEqual(2, self.machine.game.player.extra_balls)

    def test_lit_memory_false(self):
        self.mock_event('extra_ball_group_no_memory_lit')
        self.mock_event('extra_ball_group_no_memory_lit_awarded')

        # light EB, make sure it lights, but don't collect
        self.fill_troughs()
        self.start_game()
        self.post_event('start_mode1')
        self.post_event('light_eb9')
        self.assertEqual(1,
            self.machine.game.player.extra_ball_group_no_memory_num_lit)
        self.assertEventCalled('extra_ball_group_no_memory_lit')
        self.assertEventCalled('extra_ball_group_no_memory_lit_awarded')

        # drain and move to next ball, EB should be unlit
        self.mock_event('extra_ball_group_no_memory_lit')
        self.mock_event('extra_ball_group_no_memory_lit_awarded')
        self.drain_all_balls()
        self.advance_time_and_run()
        self.assertEqual(2, self.machine.game.player.ball)
        self.assertEqual(0,
            self.machine.game.player.extra_ball_group_no_memory_num_lit)
        self.assertEventNotCalled('extra_ball_group_no_memory_lit')
        self.assertEventNotCalled('extra_ball_group_no_memory_lit_awarded')



"""A switch conntected to a fast controller."""
import logging

from mpf.core.platform import SwitchConfig
from mpf.core.utility_functions import Util
from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.platforms.fast.fast import FastHardwarePlatform    # pylint: disable-msg=cyclic-import,unused-import


class FASTSwitch(SwitchPlatformInterface):

    """A switch conntected to a fast controller."""

    __slots__ = ["log", "connection", "send", "platform", "platform_settings", "_configured_debounce"]

    def __init__(self, config: SwitchConfig, number_tuple, platform: "FastHardwarePlatform", platform_settings) -> None:
        """Initialise switch."""
        super().__init__(config, number_tuple)
        self.log = logging.getLogger('FASTSwitch')
        self.connection = number_tuple[1]
        self.send = platform.net_connection.send
        self.platform = platform
        self.platform_settings = platform_settings
        self._configured_debounce = False
        self.configure_debounce(config.debounce in ("normal", "auto"))

    def get_board_name(self):
        """Return the board of this switch."""
        if self.platform.machine_type == 'wpc':
            return "FAST WPC"

        switch_index = 0
        number = Util.hex_string_to_int(self.number)
        for board_obj in self.platform.io_boards.values():
            if switch_index <= number < switch_index + board_obj.switch_count:
                return "FAST Board {}".format(str(board_obj.node_id))
            switch_index += board_obj.switch_count

        # fall back if not found
        return "FAST Unknown Board"

    def configure_debounce(self, debounce):
        """Configure debounce settings."""
        if debounce:
            debounce_open = Util.int_to_hex_string(self.platform.config['default_normal_debounce_open'])
            debounce_close = Util.int_to_hex_string(self.platform.config['default_normal_debounce_close'])
        else:
            debounce_open = Util.int_to_hex_string(self.platform.config['default_quick_debounce_open'])
            debounce_close = Util.int_to_hex_string(self.platform.config['default_quick_debounce_close'])

        if 'debounce_open' in self.platform_settings and self.platform_settings['debounce_open'] is not None:
            debounce_open = self.platform.convert_number_from_config(self.platform_settings['debounce_open'])

        if 'debounce_close' in self.platform_settings and self.platform_settings['debounce_close'] is not None:
            debounce_close = self.platform.convert_number_from_config(self.platform_settings['debounce_close'])

        if self.connection:
            cmd = 'SN:'
        else:
            cmd = 'SL:'

        new_setting = (debounce_open, debounce_close)
        if new_setting == self._configured_debounce:
            return

        self._configured_debounce = new_setting

        cmd = '{}{},01,{},{}'.format(
            cmd,
            self.number[0],
            debounce_open,
            debounce_close)

        self.send(cmd)



"""Base class for config players which have multiple entries."""
from typing import List

import abc

from mpf.core.utility_functions import Util

from mpf.core.config_player import ConfigPlayer
from mpf.exceptions.base_error import BaseError
from mpf.exceptions.config_file_error import ConfigFileError


class DeviceConfigPlayer(ConfigPlayer, metaclass=abc.ABCMeta):

    """Base class for config players which have multiple entries."""

    __slots__ = []  # type: List[str]

    allow_placeholders_in_keys = False

    def expand_config_entry(self, settings):
        """Expend objects in config entry idempotently."""
        expanded_config = dict()
        for device, device_settings in settings.items():
            device_settings = self._expand_device_config(device_settings)

            devices = self._expand_device(device)

            for this_device in devices:
                expanded_config[this_device] = device_settings

        return expanded_config

    def validate_config_entry(self, settings, name):
        """Validate one entry of this player."""
        validated_config = dict()

        # settings here is the same as a show entry, so we process with
        # that
        if not isinstance(settings, dict):
            if isinstance(settings, (str, int, float)):
                settings = self.get_string_config(settings)
            else:
                raise AssertionError(
                    "Invalid settings for player {}:{} {}".format(
                        name, self.show_section, settings))

        # settings here are dicts of devices/settings
        for device, device_settings in settings.items():
            try:
                validated_config.update(
                    self._validate_config_item(device, device_settings))
            except ConfigFileError as e:
                e.extend("Failed to load config player {}:{} with settings {}".format(
                    name, self.show_section, settings))
                raise e

        return validated_config

    # pylint: disable-msg=no-self-use
    def get_string_config(self, string):
        """Parse string config."""
        return {string: dict()}

    def _validate_config_item(self, device, device_settings):
        """Validate show config."""
        # override if you need a different show processor from config file
        # processor

        # the input values are this section's single step in a show.

        # keys are device names
        # vales are either scalars with express settings, or dicts with full
        # settings

        if device_settings is None:
            device_settings = device

        device_settings = self._parse_config(device_settings, device)
        try:
            device_settings = self._expand_device_config(device_settings)
        except BaseError as e:
            e.extend('Failed to expand entry "{}".'.format(device))
            raise e

        devices = self._expand_device(device)

        return_dict = dict()
        for this_device in devices:
            return_dict[this_device] = device_settings

        return return_dict

    def _expand_device(self, device):
        """Idempotently expand device if it is a placeholder."""
        if not isinstance(device, str):
            return [device]

        device_or_tag_names = Util.string_to_event_list(device)
        if not self.device_collection:
            return device_or_tag_names

        device_list = []
        for device_name in device_or_tag_names:
            try:
                devices = self.device_collection.items_tagged(device_name)
                if not devices:
                    device_list.append(self.device_collection[device_name])
                else:
                    device_list.extend(devices)

            except KeyError:
                if not self.__class__.allow_placeholders_in_keys or "(" not in device_name:
                    # no placeholders
                    return self.raise_config_error(
                        "Could not find a {} device with name or tag {}, from list {}".format(
                            self.device_collection.name, device_name, device_or_tag_names),
                        101)

                # placeholders may be evaluated later
                device_list.append(device_name)
        return device_list

    def _expand_device_config(self, device_settings):
        """Idempotently expand device config."""
        return device_settings

    @abc.abstractmethod
    def play(self, settings, context: str, calling_context: str, priority: int = 0, **kwargs):
        """Directly play player."""
        # **kwargs since this is an event callback
        raise NotImplementedError

    @abc.abstractmethod
    def get_express_config(self, value) -> dict:
        """Parse short config version.

        Implements "express" settings for this config_player which is what
        happens when a config is passed as a string instead of a full config
        dict. (This is detected automatically and this method is only called
        when the config is not a dict.)

        For example, the led_player uses the express config to parse a string
        like 'ff0000-f.5s' and translate it into:

        color: 220000
        fade: 500

        Since every config_player is different, this method raises a
        NotImplementedError and most be configured in the child class.

        Args:
        ----
            value: The single line string value from a config file.

        Returns a dictionary (which will then be passed through the config
        validator)
        """
        raise NotImplementedError(self.config_file_section)



"""Test shows."""
import time

from unittest.mock import MagicMock

from mpf.platforms.interfaces.driver_platform_interface import PulseSettings
from mpf.tests.MpfTestCase import MpfTestCase, test_config


class TestShows(MpfTestCase):

    def get_config_file(self):
        return 'test_shows.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/shows/'

    def get_platform(self):
        return 'smart_virtual'

    def event_handler(self, **kwargs):
        pass

    def test_default_shows(self):
        # test off
        show_off = self.machine.shows['off'].play(show_tokens=dict(leds='led_01', lights='light_01'))
        self.advance_time_and_run(.1)
        self.assertLightColor("led_01", [0, 0, 0])
        self.assertLightChannel("light_01", 0)
        show_off.stop()
        self.advance_time_and_run(.1)
        self.assertLightColor("led_01", [0, 0, 0])
        self.assertLightChannel("light_01", 0)

        # test on
        show_on = self.machine.shows['on'].play(show_tokens=dict(leds='led_01', lights='light_01'))
        self.advance_time_and_run(.1)
        self.assertLightColor("led_01", [255, 255, 255])
        self.assertLightChannel("light_01", 255)
        show_on.stop()
        self.advance_time_and_run(.1)
        self.assertLightColor("led_01", [0, 0, 0])
        self.assertLightChannel("light_01", 0)

        # test flash
        # initially on
        show_flash = self.machine.shows['flash'].play(show_tokens=dict(leds='led_01', lights='light_01'))
        self.advance_time_and_run(.1)
        self.assertLightColor("led_01", [255, 255, 255])
        self.assertLightChannel("light_01", 255)

        # after 1s off
        self.advance_time_and_run(1)
        self.assertLightColor("led_01", [0, 0, 0])
        self.assertLightChannel("light_01", 0)

        # on after another 1s
        self.advance_time_and_run(1)
        self.assertLightColor("led_01", [255, 255, 255])
        self.assertLightChannel("light_01", 255)

        show_flash.stop()
        self.advance_time_and_run(.1)
        self.assertLightColor("led_01", [0, 0, 0])
        self.assertLightChannel("light_01", 0)

        # test led_color
        show_led_color = self.machine.shows['led_color'].play(show_tokens=dict(leds='led_01', color="red"))
        self.advance_time_and_run(.1)
        self.assertLightColor("led_01", [255, 0, 0])
        show_led_color.stop()
        self.advance_time_and_run(.1)
        self.assertLightColor("led_01", [0, 0, 0])

    def test_shows(self):
        # Make sure required modes have been loaded
        self.assertIn('mode1', self.machine.modes)
        self.assertIn('mode2', self.machine.modes)
        self.assertIn('mode3', self.machine.modes)

        # Make sure test shows exist and can be loaded
        self.assertIn('test_show1', self.machine.shows)
        self.assertIn('test_show2', self.machine.shows)
        self.assertIn('test_show3', self.machine.shows)

        # Make sure hardware devices have been configured for tests
        self.assertIn('led_01', self.machine.lights)
        self.assertIn('led_02', self.machine.lights)
        self.assertIn('light_01', self.machine.lights)
        self.assertIn('light_02', self.machine.lights)
        self.assertIn('gi_01', self.machine.lights)
        self.assertIn('coil_01', self.machine.coils)
        self.assertIn('flasher_01', self.machine.lights)

        # --------------------------------------------------------
        # test_show1 - Show with LEDs, lights, and GI
        # --------------------------------------------------------

        # LEDs should start out off (current color is default RGBColor object)
        self.assertLightColor("led_01", "off")
        self.assertLightColor("led_02", "off")

        # Lights should start out off (brightness is 0)
        self.assertLightChannel("light_01", 0)
        self.assertLightChannel("light_02", 0)

        # GI should start out disabled
        self.assertLightChannel("gi_01", 0)
        self.assertEqual(self.machine.shows['test_show1'].total_steps, 5)

        # Start mode1 mode (should automatically start the test_show1 show)
        self.machine.events.post('start_mode1')
        self.advance_time_and_run(.2)
        self.assertTrue(self.machine.mode_controller.is_active('mode1'))
        self.assertTrue(self.machine.modes["mode1"].active)
        self.assertIn(self.machine.modes["mode1"],
                      self.machine.mode_controller.active_modes)

        # Grab the running show instance
        running_show1 = self.machine.show_player.instances['mode1']['show_player']['test_show1']
        self.assertIsNotNone(running_show1)

        # Make sure the show is running at the proper priority (of the mode)
        self.assertEqual(running_show1.show_config.priority, 200)

        # Check LEDs, lights, and GI after first show step
        self.assertLightColor("led_01", '006400')
        self.assertEqual(200, self.machine.lights["led_01"].stack[0].priority)
        self.assertLightColor("led_02", 'CCCCCC')
        self.assertEqual(200, self.machine.lights["led_02"].stack[0].priority)
        self.assertLightChannel("light_01", 204)
        self.assertEqual(200, self.machine.lights["light_01"].stack[0].priority)
        self.assertLightChannel("light_02", 120)
        self.assertEqual(200, self.machine.lights["light_02"].stack[0].priority)
        self.assertLightChannel("gi_01", 255)

        # Check LEDs, lights, and GI after 2nd step
        self.advance_time_and_run()
        self.assertLightColor("led_01", 'DarkGreen')
        self.assertEqual(200, self.machine.lights["led_01"].stack[0].priority)

        self.assertLightColor("led_02", 'Black')
        self.assertEqual(200, self.machine.lights["led_02"].stack[0].priority)
        self.assertLightChannel("light_01", 204)
        self.assertEqual(200, self.machine.lights["light_01"].stack[0].priority)
        self.assertLightChannel("light_02", 120)
        self.assertEqual(200, self.machine.lights["light_02"].stack[0].priority)
        self.assertLightChannel("gi_01", 255)

        # Check LEDs, lights, and GI after 3rd step
        self.advance_time_and_run()
        self.assertLightColor("led_01", 'DarkSlateGray')
        self.assertEqual(200, self.machine.lights["led_01"].stack[0].priority)
        self.assertLightColor("led_02", 'Tomato')
        self.assertEqual(200, self.machine.lights["led_02"].stack[0].priority)
        self.assertLightChannel("light_01", 255)
        self.assertEqual(200, self.machine.lights["light_01"].stack[0].priority)
        self.assertLightChannel("light_02", 51)
        self.assertEqual(200, self.machine.lights["light_02"].stack[0].priority)
        self.assertLightChannel("gi_01", 153)

        # Check LEDs, lights, and GI after 4th step (includes a fade to next
        #  color)
        self.advance_time_and_run()
        self.assertNotLightColor("led_01", 'MidnightBlue')
        self.assertEqual(200, self.machine.lights["led_01"].stack[0].priority)
        self.assertNotLightColor("led_02", 'DarkOrange')
        self.assertEqual(200, self.machine.lights["led_02"].stack[0].priority)
        self.assertLightChannel("light_01", 255)
        self.assertEqual(200, self.machine.lights["light_01"].stack[0].priority)
        self.assertLightChannel("light_02", 51)
        self.assertEqual(200, self.machine.lights["light_02"].stack[0].priority)
        self.assertLightChannel("gi_01", 51)

        # Advance time so fade should have completed
        self.advance_time_and_run(0.1)
        self.advance_time_and_run(0.1)
        self.advance_time_and_run(0.1)
        self.advance_time_and_run(0.1)
        self.advance_time_and_run(0.1)
        self.advance_time_and_run(0.1)
        self.assertLightColor("led_01", 'MidnightBlue')
        self.assertLightColor("led_02", 'DarkOrange')

        # Check LEDs after 5th step (includes a fade to black/off)
        self.advance_time_and_run(0.4)
        self.assertNotLightColor("led_01", 'Off')
        self.assertNotLightColor("led_02", 'Off')
        self.assertNotLightChannel("light_01", 0)
        self.assertNotLightChannel("light_02", 0)

        # Advance time so fade should have completed
        self.advance_time_and_run(0.2)
        self.advance_time_and_run(0.2)
        self.advance_time_and_run(0.2)
        self.advance_time_and_run(0.2)
        self.advance_time_and_run(0.2)
        self.assertLightColor("led_01", 'Off')
        self.assertLightColor("led_02", 'Off')
        self.assertLightChannel("light_01", 0)
        self.assertLightChannel("light_02", 0)
        self.assertLightChannel("gi_01", 0)

        # Make sure show loops back to the first step
        self.advance_time_and_run(1.1)
        self.assertLightColor("led_01", '006400')
        self.assertLightColor("led_02", 'CCCCCC')
        self.assertLightChannel("light_01", 204)
        self.assertLightChannel("light_02", 120)
        self.assertLightChannel("gi_01", 255)

        self.assertNotIn("show_from_mode", self.machine.show_player.instances['mode1']['show_player'])
        self.machine.variables.set_machine_var("test", 42)
        self.advance_time_and_run(.01)
        self.assertTrue(self.machine.show_player.instances['mode1']['show_player']['machine.test == 42'])

        self.machine.variables.set_machine_var("test", 23)
        self.advance_time_and_run(.01)
        self.assertTrue(self.machine.show_player.instances['mode1']['show_player']['test_key1'])
        self.assertTrue(self.machine.show_player.instances['mode1']['show_player']['test_key2'])
        self.assertNotIn("machine.test == 42", self.machine.show_player.instances['mode1']['show_player'])

        # Stop the mode (and therefore the show)
        self.machine.events.post('stop_mode1')
        self.machine_run()
        self.assertNotIn("machine.test == 42", self.machine.show_player.instances['mode1']['show_player'])
        self.assertNotIn("test_key1", self.machine.show_player.instances['mode1']['show_player'])
        self.assertNotIn("test_key2", self.machine.show_player.instances['mode1']['show_player'])
        self.assertFalse(self.machine.mode_controller.is_active('mode1'))
        self.assertTrue(running_show1._stopped)
        self.assertNotIn("test_show1", self.machine.show_player.instances['mode1']['show_player'])
        self.advance_time_and_run(5)

        # Make sure the lights and LEDs have reverted back to their prior
        # states from before the show started

        self.assertLightColor("led_01", "off")
        self.assertFalse(self.machine.lights["led_01"].stack)
        self.assertLightColor("led_01", "off")
        self.assertFalse(self.machine.lights["led_02"].stack)
        self.assertLightChannel("light_01", 0)
        self.assertFalse(self.machine.lights["light_01"].stack)
        self.assertLightChannel("light_02", 0)
        self.assertFalse(self.machine.lights["light_02"].stack)
        self.assertLightChannel("gi_01", 0)

        # --------------------------------------------------------
        # test_show2 - Show with events
        # --------------------------------------------------------

        # Setup callback for test_event event (fired in test show) and triggers
        self.event_handler = MagicMock()
        self.event_handler_2 = MagicMock()
        self.machine.events.add_handler('test_event', self.event_handler)
        self.machine.events.add_handler('play_sound', self.event_handler_2)

        # Advance the clock enough to ensure the shows have time to load
        self.assertEqual(self.machine.shows['test_show2'].total_steps, 3)

        # Make sure our event callbacks have not been fired yet
        self.assertFalse(self.event_handler.called)
        self.assertFalse(self.event_handler_2.called)

        # Start the mode that will trigger playback of the test_show2 show
        self.machine.events.post('start_mode2')
        self.machine_run()
        self.assertTrue(self.machine.mode_controller.is_active('mode2'))
        self.assertTrue(self.machine.modes["mode2"].active)
        self.assertIn(self.machine.modes["mode2"],
                      self.machine.mode_controller.active_modes)
        self.assertTrue(self.machine.show_player.instances['mode2']['show_player']['test_show2'])
        self.machine_run()

        # Make sure event callbacks have been called
        self.assertTrue(self.event_handler.called)
        self.assertTrue(self.event_handler_2.called)
        self.event_handler_2.assert_called_with(priority=0,
                                                sound="test_1",
                                                loops=-1,
                                                volume=0.5)

        # Advance to next show step and check for event
        self.advance_time_and_run()
        self.event_handler_2.assert_called_with(priority=0,
                                                sound="test_2")

        # Advance to next show step and check for event
        self.advance_time_and_run()
        self.event_handler_2.assert_called_with(priority=0,
                                                sound="test_3",
                                                loops=1,
                                                volume=0.35)

        # Stop the mode (and therefore the show)
        self.machine.events.post('stop_mode2')
        self.machine_run()
        self.assertFalse(self.machine.mode_controller.is_active('mode2'))
        self.advance_time_and_run(5)

        # --------------------------------------------------------
        # test_show3 - Show with coils and flashers
        # --------------------------------------------------------

        # Setup callback for test_event event (fired in test show) and triggers
        self.machine.coils['coil_01'].hw_driver.pulse = MagicMock()
        self.machine.coils['flasher_01'].hw_driver.enable = MagicMock()

        self.assertEqual(self.machine.shows['test_show3'].total_steps, 3)

        # Make sure our device callbacks have not been fired yet
        self.assertFalse(self.machine.coils['coil_01'].hw_driver.pulse.called)
        self.assertFalse(self.machine.coils['flasher_01'].hw_driver.enable.called)

        # Start the mode that will trigger playback of the test_show3 show
        self.machine.events.post('start_mode3')
        self.machine_run()
        self.assertTrue(self.machine.mode_controller.is_active('mode3'))
        self.assertTrue(self.machine.modes["mode3"].active)
        self.assertIn(self.machine.modes["mode3"],
                      self.machine.mode_controller.active_modes)
        self.assertTrue(self.machine.show_player.instances['mode3']['show_player']['test_show3'])
        self.machine_run()

        # Make sure flasher device callback has been called (in first step
        # of show)
        self.advance_time_and_run(.01)
        self.assertTrue(self.machine.coils['flasher_01'].hw_driver.enable.called)

        # Advance to next show step and check for coil firing
        self.advance_time_and_run()
        self.machine.coils['coil_01'].hw_driver.pulse.assert_called_with(PulseSettings(power=1.0, duration=30))

        # Advance to next show step and check for coil firing
        self.advance_time_and_run()
        self.machine.coils['coil_01'].hw_driver.pulse.assert_called_with(PulseSettings(power=0.45, duration=30))

        # TODO: Test device tags
        # TODO: Add test for multiple shows running at once with different
        # priorities
        # TODO: Test show playback rate

    def test_duration_in_shows(self):
        show = self.machine.shows['show_with_time_and_duration']
        self.assertEqual(6, len(show.show_steps))
        self.assertEqual(1, show.show_steps[0]['duration'])
        self.assertEqual(4, show.show_steps[1]['duration'])
        self.assertEqual(1, show.show_steps[2]['duration'])
        self.assertEqual(1, show.show_steps[3]['duration'])
        self.assertEqual(3, show.show_steps[4]['duration'])
        self.assertEqual(3, show.show_steps[5]['duration'])

    def test_tokens_in_shows(self):
        self.assertIn('leds_name_token', self.machine.shows)
        self.assertIn('leds_color_token', self.machine.shows)
        self.assertIn('leds_color_token_and_fade', self.machine.shows)
        self.assertIn('leds_extended', self.machine.shows)
        self.assertIn('lights_basic', self.machine.shows)
        self.assertIn('multiple_tokens', self.machine.shows)

        # test keys passed via method calls

        # test one LED
        show = self.machine.shows['leds_name_token'].play(
            show_tokens=dict(leds='led_01'))
        self.advance_time_and_run(.5)

        self.assertLightColor("led_01", 'red')
        show.stop()

        # test passing tag instead of LED name
        self.machine.lights["led_01"].clear_stack()
        self.advance_time_and_run()

        show = self.machine.shows['leds_name_token'].play(show_tokens=dict(
            leds='tag1'))
        self.advance_time_and_run(.5)
        self.assertLightColor("led_01", 'red')
        self.assertLightColor("led_02", 'red')
        show.stop()

        # test passing multiple LEDs as string list
        self.machine.lights["led_01"].clear_stack()
        self.advance_time_and_run()

        show = self.machine.shows['leds_name_token'].play(show_tokens=dict(leds='led_01, led_02'))
        self.advance_time_and_run(.5)
        self.assertLightColor("led_01", 'red')
        self.assertLightColor("led_02", 'red')
        show.stop()

        # test passing color as a token
        self.machine.lights["led_01"].clear_stack()
        self.machine.lights["led_02"].clear_stack()
        self.advance_time_and_run()

        show = self.machine.shows['leds_color_token'].play(
            show_tokens=dict(color1='blue', color2='green'))
        self.advance_time_and_run(2)
        self.assertLightColor("led_01", 'blue')
        self.assertLightColor("led_02", 'green')

        show.stop()

        # Test passing color as a token and include a fade string in express config
        show = self.machine.shows['leds_color_token_and_fade'].play(
            show_tokens=dict(color1='blue', color2='green'))
        self.advance_time_and_run(2)
        self.assertLightColor("led_01", 'blue')
        self.assertLightColor("led_02", 'green')

        show.stop()

        # test single LED in show with extended LED config
        self.machine.lights["led_01"].clear_stack()
        self.machine.lights["led_02"].clear_stack()
        self.advance_time_and_run()

        # led should be off
        self.assertLightColor("led_01", [0, 0, 0])

        show = self.machine.shows['leds_extended'].play(
            show_tokens=dict(leds='led_01'))
        self.advance_time_and_run(.5)

        # show has fade of 1s, so after 0.5s it should be halfway to red
        # we use assertColorAlmostEqual because timing of shows + led fades is not stable in tests
        #self.assertColorAlmostEqual(self.machine.leds.led_01.hw_driver.current_color, [127, 0, 0])
        show.stop()

        # test tag in show with extended LED config
        self.machine.lights["led_01"].clear_stack()
        self.machine.lights["led_02"].clear_stack()
        self.advance_time_and_run()

        show = self.machine.shows['leds_extended'].play(
            show_tokens=dict(leds='tag1'))
        self.advance_time_and_run(.51)

        # show has fade of 1s, so after 0.5s it should be halfway to red
        # we use assertColorAlmostEqual because timing of shows + led fades is not stable in tests
        #self.assertColorAlmostEqual(self.machine.leds.led_01.hw_driver.current_color, [127, 0, 0])
        #self.assertColorAlmostEqual(self.machine.leds.led_02.hw_driver.current_color, [127, 0, 0])
        show.stop()

        # test single light in show
        show = self.machine.shows['lights_basic'].play(
            show_tokens=dict(lights='light_01'))
        self.advance_time_and_run(.5)

        self.assertLightChannel("light_01", 255)
        show.stop()

        # test light tag in show
        self.machine.lights["light_01"].off(force=True)
        self.advance_time_and_run()

        show = self.machine.shows['lights_basic'].play(
            show_tokens=dict(lights='tag1'))
        self.advance_time_and_run(.5)

        self.assertLightChannel("light_01", 255)
        self.assertLightChannel("light_02", 255)
        show.stop()

        # test lights as string list in show
        self.machine.lights["light_01"].off(force=True)
        self.advance_time_and_run()

        show = self.machine.shows['lights_basic'].play(
            show_tokens=dict(lights='light_01, light_02'))
        self.advance_time_and_run(.5)

        self.assertLightChannel("light_01", 255)
        self.assertLightChannel("light_02", 255)
        show.stop()

        # test led and light tags in the same show
        self.machine.lights["led_01"].clear_stack()
        self.machine.lights["led_02"].clear_stack()
        self.machine.lights["light_01"].off(force=True)
        self.machine.lights["light_02"].off(force=True)
        self.advance_time_and_run()

        show = self.machine.shows['multiple_tokens'].play(
            show_tokens=dict(leds='led_01', lights='light_01'))
        self.advance_time_and_run(.5)
        self.assertLightColor("led_01", 'blue')
        self.assertLightChannel("light_01", 255)
        show.stop()
        self.advance_time_and_run(.5)

        self.assertLightColor("led_01", 'black')
        self.assertLightColor("led_02", 'black')

        self.post_event("play_show_assoc_tokens")
        self.advance_time_and_run(.5)

        self.assertLightColor("led_01", 'red')
        self.assertLightColor("led_02", 'red')

        self.post_event("stop_show_assoc_tokens")
        self.advance_time_and_run(.5)

        self.assertLightColor("led_01", 'black')
        self.assertLightColor("led_02", 'black')

        self.post_event("test_mode_started")
        self.advance_time_and_run(.5)

        self.assertLightColor("led_01", 'red')
        self.post_event("test_mode_stopped")

    def test_get_show_copy(self):
        copied_show = self.machine.shows['test_show1'].get_show_steps()
        self.assertEqual(5, len(copied_show))
        self.assertIs(type(copied_show), list)
        self.assertEqual(copied_show[0]['duration'], 1.0)
        self.assertEqual(copied_show[1]['duration'], 1.0)
        self.assertEqual(copied_show[2]['duration'], 1.0)
        self.assertEqual(copied_show[4]['duration'], 2.0)
        self.assertIn(self.machine.lights["led_01"], copied_show[0]['lights'])
        self.assertIn(self.machine.lights["led_02"], copied_show[0]['lights'])
        self.assertEqual(copied_show[0]['lights'][self.machine.lights["led_01"]],
                         dict(color='006400', fade=None, priority=0))
        self.assertEqual(copied_show[0]['lights'][self.machine.lights["led_02"]],
                         dict(color='cccccc', fade=None, priority=0))
        self.assertEqual(copied_show[3]['lights'][self.machine.lights["led_01"]],
                         dict(color='midnightblue', fade=500, priority=0))

    def test_show_player(self):
        # Basic show
        self.machine.events.post('play_test_show1')
        self.advance_time_and_run()
        self.assertEqual(1, len(self.machine.show_player.instances['_global']['show_player']))
        self.post_event("stop_test_show1")

        # Test priority
        self.machine.events.post('play_with_priority')
        self.advance_time_and_run()
        self.assertEqual(15, self.machine.show_player.instances['_global']['show_player']['test_show1'].show_config.priority)
        self.post_event("stop_test_show1")

        # Test speed
        self.machine.events.post('play_with_speed')
        self.advance_time_and_run()
        self.assertEqual(2, self.machine.show_player.instances['_global']['show_player']['test_show1'].show_config.speed)
        self.post_event("stop_test_show1")

        # Test start step
        self.machine.events.post('play_with_start_step')
        self.advance_time_and_run(.1)
        self.assertEqual(2, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_index)
        self.post_event("stop_test_show1")

        # Test start step
        self.machine.events.post('play_with_neg_start_step')
        self.advance_time_and_run(.1)
        self.assertEqual(4, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_index)
        self.post_event("stop_test_show1")

        # Test loops
        self.machine.events.post('play_with_loops')
        self.advance_time_and_run(.1)
        self.assertEqual(2, self.machine.show_player.instances['_global']['show_player']['test_show1'].loops)
        self.post_event("stop_test_show1")

        # Test sync_ms 1000ms
        self.machine.events.post('play_with_sync_ms_1000')
        self.advance_time_and_run(.1)

        # should be 0 +/- the duration of a frame

        self.assertAlmostEqual(0.0, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_time % 1.0, delta=(1 / 30))
        self.post_event("stop_test_show1")

        # Test sync_ms 500ms
        self.machine.events.post('play_with_sync_ms_500')
        self.advance_time_and_run(.1)
        self.assertAlmostEqual(0.0, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_time % 0.5, delta=(1 / 30))
        self.post_event("stop_test_show1")

        # Test manual advance
        self.machine.events.post('play_with_manual_advance')
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_index)

    def advance_to_sync_ms(self, ms):
        current_time = self.clock.get_time()
        ms = float(ms)
        next_full_second = current_time + ((ms - (current_time * 1000) % ms) / 1000.0)
        self.advance_time_and_run(next_full_second - current_time)
        delta = (self.clock.get_time() * 1000) % ms
        self.assertTrue(delta == ms or delta < 1, "Delta {} too large".format(delta))

    @test_config("test_sync_ms.yaml")
    def test_sync_ms(self):
        self.advance_to_sync_ms(250)
        # shortly after sync point. start initial show
        self.advance_time_and_run(.01)
        self.post_event("play_show_sync_ms1")
        self.assertLightColor("light", "off")
        # wait for first sync point
        self.advance_to_sync_ms(250)
        self.advance_time_and_run(.01)
        self.assertLightColor("light", "red")
        # shortly after sync point again
        self.post_event("play_show_sync_ms2")
        # old show still playing
        self.advance_time_and_run(.1)
        self.assertLightColor("light", "red")
        # until next sync point is reached
        self.advance_time_and_run(.15)
        self.assertLightColor("light", "blue")
        self.post_event("stop_show")
        self.assertLightColor("light", "off")

        # play show and start second show before first was synced
        self.advance_to_sync_ms(250)
        self.advance_time_and_run(.01)
        self.post_event("play_show_sync_ms1")
        self.advance_time_and_run(.01)
        self.post_event("play_show_sync_ms2")
        self.assertLightColor("light", "off")
        self.advance_to_sync_ms(250)
        self.assertLightColor("light", "blue")
        self.post_event("stop_show")
        self.assertLightColor("light", "off")

        # play show and start second show before first was synced. stop before second is synced
        self.advance_to_sync_ms(250)
        self.advance_time_and_run(.01)
        self.post_event("play_show_sync_ms1")
        self.advance_time_and_run(.01)
        self.post_event("play_show_sync_ms2")
        self.assertLightColor("light", "off")
        self.post_event("stop_show")
        self.assertLightColor("light", "off")
        self.advance_time_and_run(.5)
        self.assertLightColor("light", "off")

    def test_pause_resume_shows(self):
        self.machine.events.post('play_test_show1')
        # make sure show is advancing
        self.advance_time_and_run(1)
        self.assertEqual(2, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_index)
        self.advance_time_and_run(1)
        self.assertEqual(3, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_index)

        self.machine.events.post('pause_test_show1')

        # make sure show stops advancing
        self.advance_time_and_run(1)
        self.assertEqual(3, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_index)
        self.advance_time_and_run(1)
        self.assertEqual(3, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_index)
        self.advance_time_and_run(1)
        self.assertEqual(3, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_index)

        # make sure show starts advanving again
        self.machine.events.post('resume_test_show1')
        self.advance_time_and_run(0.1)
        self.assertEqual(4, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_index)
        self.advance_time_and_run(1)
        self.assertEqual(5, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_index)
        self.advance_time_and_run(2)
        self.assertEqual(1, self.machine.show_player.instances['_global']['show_player']['test_show1'].next_step_index)

    def test_show_from_mode_config(self):
        self.assertIn('show_from_mode', self.machine.shows)

        self.assertEqual(1, len(self.machine.shows['show_from_mode'].show_steps))

    def test_too_many_tokens(self):
        with self.assertRaises(ValueError):
            self.machine.shows['lights_basic'].play(show_tokens=dict(
                lights='light_01', fake='foo'))

    def test_too_few_tokens(self):
        self.machine.shows['multiple_tokens'].play(show_tokens=dict(
            lights='light_01'))

    def test_keys_in_show_player(self):
        self.post_event("play_on_led1")
        self.advance_time_and_run()
        self.assertLightColor("led_01", [255, 255, 255])
        self.assertLightColor("led_02", [0, 0, 0])

        self.post_event("play_on_led2")
        self.advance_time_and_run()
        self.assertLightColor("led_01", [255, 255, 255])
        self.assertLightColor("led_02", [255, 255, 255])

        self.post_event("stop_on_led1")
        self.advance_time_and_run()
        self.assertLightColor("led_01", [0, 0, 0])
        self.assertLightColor("led_02", [255, 255, 255])

        self.post_event("stop_on_led2")
        self.advance_time_and_run()
        self.assertLightColor("led_01", [0, 0, 0])
        self.assertLightColor("led_02", [0, 0, 0])

    def test_nested_shows(self):
        self.mock_event("test")
        show = self.machine.shows['myparentshow'].play(loops=0)

        self.advance_time_and_run(1)
        self.assertEqual(1, self._events['test'])
        show.stop()

    def test_manual_advance(self):
        self.assertLightColor("led_01", [0, 0, 0])
        self.post_event("play_manual_advance")
        self.advance_time_and_run()
        self.assertLightColor("led_01", [255, 0, 0])

        self.post_event("advance_manual_advance")
        self.advance_time_and_run()
        self.assertLightColor("led_01", [0, 255, 0])

        self.post_event("advance_manual_advance")
        self.advance_time_and_run()
        self.assertLightColor("led_01", [0, 0, 255])

        self.post_event("advance_manual_step_back")
        self.advance_time_and_run()
        self.assertLightColor("led_01", [0, 255, 0])

        self.post_event("advance_manual_advance")
        self.advance_time_and_run()
        self.assertLightColor("led_01", [0, 0, 255])

        self.post_event("play_manual_advance")
        self.advance_time_and_run()
        self.assertLightColor("led_01", [255, 0, 0])

        # test wrap around
        self.post_event("advance_manual_step_back")
        self.advance_time_and_run()
        self.assertLightColor("led_01", [0, 0, 255])

    def test_queue_event(self):
        done = MagicMock()
        self.mock_event("step1")
        self.mock_event("step2")
        self.mock_event("step3")
        self.machine.events.post_queue("queue_play", done)
        self.advance_time_and_run(.1)
        self.assertEventCalled("step1")
        self.assertEventNotCalled("step2")
        self.assertEventNotCalled("step3")
        done.assert_not_called()
        self.advance_time_and_run(1)
        self.assertEventCalled("step1")
        self.assertEventCalled("step2")
        self.assertEventNotCalled("step3")
        done.assert_not_called()
        self.advance_time_and_run(1)
        self.assertEventCalled("step1")
        self.assertEventCalled("step2")
        self.assertEventCalled("step3")
        done.assert_not_called()
        self.mock_event("step1")
        self.mock_event("step2")
        self.mock_event("step3")
        self.advance_time_and_run(1)
        self.assertEventNotCalled("step1")
        self.assertEventNotCalled("step2")
        self.assertEventNotCalled("step3")
        done.assert_called_once_with()

    def test_show_player_emitted_events(self):
        self.mock_event('test_show1_played')
        self.mock_event('test_show1_played2')
        self.mock_event('test_show1_stopped')
        self.mock_event('test_show1_looped')
        self.mock_event('test_show1_paused')
        self.mock_event('test_show1_resumed')
        self.mock_event('test_show1_advanced')
        self.mock_event('test_show1_stepped_back')
        self.mock_event('test_show1_completed')

        self.machine.events.post('play_with_emitted_events')
        self.advance_time_and_run(1)
        self.assertEventCalled('test_show1_played')
        self.assertEventCalled('test_show1_played2')
        self.assertEventNotCalled('test_show1_stopped')
        self.assertEventNotCalled('test_show1_looped')
        self.assertEventNotCalled('test_show1_paused')
        self.assertEventNotCalled('test_show1_resumed')
        self.assertEventNotCalled('test_show1_advanced')
        self.assertEventNotCalled('test_show1_stepped_back')
        self.assertEventNotCalled('test_show1_completed')

        self.advance_time_and_run(6)
        self.assertEventCalled('test_show1_looped')
        self.assertEventNotCalled('test_show1_stopped')
        self.assertEventNotCalled('test_show1_paused')
        self.assertEventNotCalled('test_show1_resumed')
        self.assertEventNotCalled('test_show1_advanced')
        self.assertEventNotCalled('test_show1_stepped_back')

        self.machine.events.post('pause_emitted_events_show')
        self.advance_time_and_run(1)
        self.assertEventCalled('test_show1_paused')

        self.machine.events.post('advance_emitted_events_show')
        self.advance_time_and_run(1)
        self.assertEventCalled('test_show1_advanced')

        self.machine.events.post('step_back_emitted_events_show')
        self.advance_time_and_run(1)
        self.assertEventCalled('test_show1_stepped_back')

        self.machine.events.post('resume_emitted_events_show')
        self.advance_time_and_run(1)
        self.assertEventCalled('test_show1_resumed')

        self.machine.events.post('stop_emitted_events_show')
        self.advance_time_and_run(1)
        self.assertEventCalled('test_show1_stopped')

        self.assertEventNotCalled('test_show1_completed')

    def test_show_player_completed_events(self):
        self.mock_event('test_show1_completed')
        self.mock_event('test_show1_stopped')

        self.machine.events.post('play_with_completed_event')
        self.advance_time_and_run(1)
        self.assertEventNotCalled('test_show1_completed')
        self.assertEventNotCalled('test_show1_stopped')

        self.advance_time_and_run(6)
        self.assertEventCalled('test_show1_completed')
        self.assertEventCalled('test_show1_stopped')

    def test_token_in_keys(self):
        self.post_event("play_show_with_token_in_key")
        self.advance_time_and_run()
        self.assertLightColor("led_01", "red")

    def test_placeholder_in_token(self):
        self.assertNotLightColor("led_02", "blue")
        self.machine.variables.set_machine_var("test_color", "blue")
        self.machine.variables.set_machine_var("test_num", "02")
        self.post_event("play_show_with_placeholder_in_token")
        self.advance_time_and_run()
        self.assertLightColor("led_02", "blue")

    def test_placeholder_in_token_and_events(self):
        self.assertNotLightColor("led_02", "red")
        self.post_event_with_params("play_show_with_placeholder_in_token_and_event_args", test_color="red",
                                    test_num="02")
        self.advance_time_and_run()
        self.assertLightColor("led_02", "red")

    def test_non_string_token(self):
        self.start_mode("mode4")
        self.assertLightColor("led_01", "black")
        self.post_event("test_token")
        self.advance_time_and_run(.05)
        self.assertNotLightColor("led_01", "red")
        self.advance_time_and_run(.06)
        self.assertLightColor("led_01", "red")

    def test_conditional_shows(self):
        self.assertLightColor("led_01", "black")
        self.post_event_with_params("play_show_with_condition_in_event", green=False)
        self.advance_time_and_run()
        self.assertLightColor("led_01", "purple")
        self.post_event_with_params("play_show_with_condition_in_event", green=True)
        self.advance_time_and_run()
        self.assertLightColor("led_01", "green")

    def test_conditional_show_names(self):
        self.assertLightColor("led_01", "black")
        self.post_event_with_params("play_show_with_condition_in_show", blue=False)
        self.advance_time_and_run()
        self.assertLightColor("led_01", "red")
        self.post_event_with_params("play_show_with_condition_in_show", blue=True)
        self.advance_time_and_run()
        self.assertLightColor("led_01", "blue")

    # Regression test for bug in standalone variable shows
    def test_variable_show(self):
        variable_show = self.machine.shows['test_variable_show'].play()
        self.advance_time_and_run(.1)
        # check that machine variable was set
        self.assertEqual(self.machine.variables.get_machine_var('foo'), 0)
        # advance to next show step
        self.advance_time_and_run()
        # check that machine variable was incremented
        self.assertEqual(self.machine.variables.get_machine_var('foo'), 1)

    @test_config("test_show_player_queue.yaml")
    def test_show_player_queue(self):
        for show_num in range(1, 4):
            for step_num in range(1, 4):
                self.mock_event("step{}_{}".format(show_num, step_num))

        self.post_event("play_show1_on_queue1")
        self.post_event("play_show2_on_queue1")

        self.advance_time_and_run(.5)
        self.assertEventCalled("step1_1")
        self.assertEventNotCalled("step1_2")
        self.assertEventNotCalled("step2_1")
        self.advance_time_and_run()
        self.assertEventCalled("step1_2")
        self.assertEventNotCalled("step1_3")
        self.assertEventNotCalled("step2_1")
        self.advance_time_and_run()
        self.assertEventCalled("step1_3")
        self.assertEventNotCalled("step2_1")
        self.advance_time_and_run()
        self.assertEventCalled("step2_1")
        self.assertEventNotCalled("step2_2")

        self.mock_event("step1_1")
        self.mock_event("step1_2")
        self.mock_event("step1_3")
        self.post_event("play_show1_on_queue2")
        self.advance_time_and_run(.1)
        self.assertEventCalled("step1_1")
        self.assertEventNotCalled("step1_2")

        self.advance_time_and_run()
        self.assertEventCalled("step2_2")
        self.assertEventNotCalled("step2_3")
        self.assertEventCalled("step1_2")
        self.assertEventNotCalled("step1_3")
